# coding: utf8
import os
import sys
import platform
from subprocess import Popen
import atexit
import signal
from ctypes import cdll

from .utils import printf, json_dumps


CWD = os.path.dirname(os.path.abspath(__file__))


# https://stackoverflow.com/a/19448255/10517783
def kill_child(proc):
    proc.kill()
    # A None value indicates that the process hasn’t terminated yet.
    # A negative value -N indicates that the child was terminated by signal N (Unix only).
    printf('Caching subprocess (pid: %s) killed with returncode: %s' % (proc.pid, proc.wait()), warn=True)


# https://stackoverflow.com/a/13256908/10517783
# https://stackoverflow.com/a/23587108/10517783
# http://evans.io/legacy/posts/killing-child-processes-on-parent-exit-prctl/
class PrCtlError(Exception):
    pass


def on_parent_exit(signame):
    """
    Return a function to be run in a child process which will trigger
    SIGNAME to be sent when the parent process dies
    """
    # On Windows, signal() can only be called with SIGABRT, SIGFPE, SIGILL, SIGINT, SIGSEGV, or SIGTERM.
    signum = getattr(signal, signame)  # SIGTERM 15  SIGKILL 9

    def set_parent_exit_signal():
        # Constant taken from http://linux.die.net/include/linux/prctl.h
        PR_SET_PDEATHSIG = 1
        # http://linux.die.net/man/2/prctl
        result = cdll['libc.so.6'].prctl(PR_SET_PDEATHSIG, signum)
        if result != 0:
            raise PrCtlError('prctl failed with error code %s' % result)

    return set_parent_exit_signal


def init_caching(config, main_pid):
    if config.get('ENABLE_CACHE', True):
        caching_subprocess = start_caching(config, main_pid)
        caching_pid = caching_subprocess.pid
        printf("Caching HTML for Log and Stats page in the background with pid: %s" % caching_pid)
        atexit.register(kill_child, caching_subprocess)
    else:
        caching_pid = None

    return caching_pid


def start_caching(config, main_pid):
    _bind = config.get('SCRAPYDWEB_BIND', '0.0.0.0')
    _bind = '127.0.0.1' if _bind == '0.0.0.0' else _bind
    args = [
        sys.executable,
        os.path.join(CWD, 'cache.py'),

        str(main_pid),
        _bind,
        str(config.get('SCRAPYDWEB_PORT', 5000)),
        config.get('USERNAME', '') if config.get('ENABLE_AUTH', False) else '',
        config.get('PASSWORD', '') if config.get('ENABLE_AUTH', False) else '',
        json_dumps(config.get('SCRAPYD_SERVERS', ['127.0.0.1'])),
        json_dumps(config.get('SCRAPYD_SERVERS_AUTHS', [None])),
        str(config.get('CACHE_ROUND_INTERVAL', 300)),
        str(config.get('CACHE_REQUEST_INTERVAL', 10)),
        str(config.get('VERBOSE', False))
    ]

    # 'Windows':
    # AttributeError: module 'signal' has no attribute 'SIGKILL'
    # ValueError: preexec_fn is not supported on Windows platforms
    # macOS('Darwin'):
    # subprocess.SubprocessError: Exception occurred in preexec_fn.
    # OSError: dlopen(libc.so.6, 6): image not found
    if platform.system() == 'Linux':
        kwargs = dict(preexec_fn=on_parent_exit('SIGKILL'))  # 'SIGTERM' 'SIGKILL'
        try:
            caching_subprocess = Popen(args, **kwargs)
        except Exception as err:
            printf(err, warn=True)
            caching_subprocess = Popen(args)
    else:
        caching_subprocess = Popen(args)

    return caching_subprocess



# coding: utf8
import os
import re
from collections import defaultdict, OrderedDict

from flask import render_template

from ..myview import MyView
from ..vars import EMAIL_TRIGGER_KEYS
from ..utils.utils import json_dumps


CWD = os.path.dirname(os.path.abspath(__file__))


class SettingsView(MyView):
    methods = ['GET']

    def __init__(self):
        super(self.__class__, self).__init__()

        self.template = 'scrapydweb/settings.html'
        self.kwargs = dict(node=self.node)

    def dispatch_request(self, **kwargs):
        self.update_kwargs()
        return render_template(self.template, **self.kwargs)

    @staticmethod
    def protect(string):
        if not isinstance(string, str):
            return string
        length = len(string)
        if length == 0:
            return string
        elif length < 3:
            return re.sub(r'^.', '*', string)
        elif length < 7:
            return re.sub(r'^.(.*?).$', r'*\1*', string)
        else:
            return re.sub(r'^..(.*?)..$', r'**\1**', string)

    @staticmethod
    def json_dumps(obj, sort_keys=False):
        string = json_dumps(obj, sort_keys=sort_keys)
        return string.replace(' true', ' True').replace(' false', ' False').replace(' null', ' None')

    def update_kwargs(self):
        # User settings
        self.kwargs['default_settings_py'] = os.path.join(os.path.dirname(CWD), 'default_settings.py')

        # ScrapydWeb
        self.kwargs['scrapydweb_server'] = self.json_dumps(dict(
            SCRAPYDWEB_BIND=self.SCRAPYDWEB_BIND,
            SCRAPYDWEB_PORT=self.SCRAPYDWEB_PORT,
            ENABLE_AUTH=self.ENABLE_AUTH,
            USERNAME=self.protect(self.USERNAME),
            PASSWORD=self.protect(self.PASSWORD)
        ))

        # Scrapy
        self.kwargs['SCRAPY_PROJECTS_DIR'] = self.SCRAPY_PROJECTS_DIR or "''"

        # Scrapyd
        servers = defaultdict(list)
        for group, server, auth in zip(self.SCRAPYD_SERVERS_GROUPS, self.SCRAPYD_SERVERS, self.SCRAPYD_SERVERS_AUTHS):
            _server = '%s:%s@%s' % (self.protect(auth[0]), self.protect(auth[1]), server) if auth else server
            servers[group].append(_server)

        self.kwargs['servers'] = self.json_dumps(servers)
        self.kwargs['SCRAPYD_LOGS_DIR'] = self.SCRAPYD_LOGS_DIR or "''"
        self.kwargs['SCRAPYD_LOG_EXTENSIONS'] = self.SCRAPYD_LOG_EXTENSIONS

        # Page Display
        self.kwargs['page_display_details'] = self.json_dumps(dict(
            SHOW_SCRAPYD_ITEMS=self.SHOW_SCRAPYD_ITEMS,
            SHOW_DASHBOARD_JOB_COLUMN=self.SHOW_DASHBOARD_JOB_COLUMN,
            DASHBOARD_RELOAD_INTERVAL=self.DASHBOARD_RELOAD_INTERVAL,
            DAEMONSTATUS_REFRESH_INTERVAL=self.DAEMONSTATUS_REFRESH_INTERVAL
        ))

        # HTML Caching
        self.kwargs['ENABLE_CACHE'] = self.ENABLE_CACHE
        self.kwargs['html_caching_details'] = self.json_dumps(dict(
            CACHE_ROUND_INTERVAL=self.CACHE_ROUND_INTERVAL,
            CACHE_REQUEST_INTERVAL=self.CACHE_REQUEST_INTERVAL,
            DELETE_CACHE=self.DELETE_CACHE
        ))

        # Email Notice
        self.kwargs['ENABLE_EMAIL'] = self.ENABLE_EMAIL
        self.kwargs['smtp_settings'] = self.json_dumps(dict(
            SMTP_SERVER=self.SMTP_SERVER,
            SMTP_PORT=self.SMTP_PORT,
            SMTP_OVER_SSL=self.SMTP_OVER_SSL,
            SMTP_CONNECTION_TIMEOUT=self.SMTP_CONNECTION_TIMEOUT
        ))

        self.kwargs['sender_recipients'] = self.json_dumps(dict(
            FROM_ADDR=self.FROM_ADDR,
            EMAIL_PASSWORD=self.protect(self.EMAIL_PASSWORD),
            TO_ADDRS=self.TO_ADDRS
        ))

        self.kwargs['email_working_time'] = self.json_dumps([
            dict(
                EMAIL_WORKING_DAYS="%s" % sorted(self.EMAIL_WORKING_DAYS),  # stringify making it displayed in a line
                remark="Monday is 1 and Sunday is 7"
            ),
            dict(
                EMAIL_WORKING_HOURS="%s" % sorted(self.EMAIL_WORKING_HOURS),
                remark="From 0 to 23"
            )
        ])

        # email triggers
        d = OrderedDict()
        d['ON_JOB_RUNNING_INTERVAL'] = self.ON_JOB_RUNNING_INTERVAL
        d['ON_JOB_FINISHED'] = self.ON_JOB_FINISHED

        for key in EMAIL_TRIGGER_KEYS:
            keys = ['LOG_%s_THRESHOLD' % key, 'LOG_%s_TRIGGER_STOP' % key, 'LOG_%s_TRIGGER_FORCESTOP' % key]
            d[key] = {k: getattr(self, k) for k in keys}
        value = self.json_dumps(d)
        value = re.sub(r'True', "<b style='color: red'>True</b>", value)
        value = re.sub(r'(\s[1-9]\d*)', r"<b style='color: red'>\1</b>", value)
        self.kwargs['email_triggers'] = value

        # System
        self.kwargs['DEBUG'] = self.DEBUG
        self.kwargs['VERBOSE'] = self.VERBOSE



# coding: utf8
import time
import datetime
import re


pattern_datetime = r'\d{4}-\d{2}-\d{2}\s\d{2}:\d{2}:\d{2}'
# 2018-06-25 18:12:49 [scrapy.extensions.logstats] INFO:
# Crawled 2318 pages (at 2 pages/min), scraped 68438 items (at 60 items/min)
pattern_datas = re.compile(r"""\n(?P<time_>%s).*?
                            Crawled\s(?P<pages>\d+)\s+pages\s
                            \(at\s(?P<pages_min>\d+)\spages/min\),\s
                            scraped\s(?P<items>\d+)\sitems\s
                            \(at\s(?P<items_min>\d+)\sitems/min\)
                            """ % pattern_datetime, re.X)

patterns = [
    r'\]\sCRITICAL:',
    r'\]\sERROR:',
    r'\]\sWARNING:',
    r':\sRedirecting\s\(',
    r':\sRetrying\s<',
    r':\sIgnoring\sresponse\s<',
]
pattern_logs_count_list = []
for p in patterns:
    pattern = re.compile(r"""\n
                            (\d{4}[^\n]*?%s.*?)
                            (?=\n\d{4}[^\n]*?(?:DEBUG|INFO|WARNING|ERROR|CRITICAL))
                        """ % p, re.S | re.X)
    pattern_logs_count_list.append(pattern)


def parse_log(text, kwargs):
    # fake_time = '2018-01-01 00:00:01'
    fake_time = time.strftime('%Y-%m-%d %H:%M:%S')
    lines = re.split('\r*\n', text)

    def re_search_final_match(pattern, default='', step=-1):
        for line in lines[::step]:
            m = re.search(pattern, line)
            if m:
                return line
        else:
            return default

    kwargs['first_log_time'] = re_search_final_match(r'^%s' % pattern_datetime, step=1)[:19] or fake_time
    kwargs['latest_log_time'] = re_search_final_match(r'^%s' % pattern_datetime)[:19] or fake_time
    first_log_datetime = datetime.datetime.strptime(kwargs['first_log_time'], '%Y-%m-%d %H:%M:%S')
    latest_log_datetime = datetime.datetime.strptime(kwargs['latest_log_time'], '%Y-%m-%d %H:%M:%S')
    kwargs['latest_log_timestamp'] = time.mktime(latest_log_datetime.timetuple())
    kwargs['elasped'] = str(latest_log_datetime - first_log_datetime)

    kwargs['head_lines'] = '\n'.join(lines[:50])
    kwargs['tail_lines'] = '\n'.join(lines[-50:])

    # Extract datas for chart
    datas = pattern_datas.findall(text)
    # For compatibility with Python 2, str(time_) to avoid [u'2018-08-22 18:43:05', 0, 0, 0, 0] in JavaScript
    kwargs['datas'] = [[str(time_), int(pages), int(pages_min), int(items), int(items_min)]
                       for (time_, pages, pages_min, items, items_min) in datas]
    kwargs['crawled_pages'] = int(datas[-1][1]) if datas else 0
    kwargs['scraped_items'] = int(datas[-1][3]) if datas else 0

    # Extract only last log
    latest_tuples = [
        # ('resuming_crawl', 'Resuming\scrawl'),
        ('latest_offsite', r'Filtered\soffsite'),
        ('latest_duplicate', r'Filtered\sduplicate'),
        ('latest_crawl', r'Crawled\s\(\d+'),
        ('latest_scrape', r'Scraped\sfrom\s'),
        ('latest_item', r'^\{.*\}'),
        ('latest_stat', r'Crawled\s\d+\spages'),
    ]
    latest_matches = [('resuming_crawl', re_search_final_match(r'Resuming\scrawl', step=1))]
    for k, v in latest_tuples:
        ret = re_search_final_match(v)
        latest_matches.append((k, ret))
        if k == 'latest_crawl':
            latest_crawl_datetime = datetime.datetime.strptime(ret[:19] or fake_time, '%Y-%m-%d %H:%M:%S')
        elif k == 'latest_scrape':
            latest_scrape_datetime = datetime.datetime.strptime(ret[:19] or fake_time, '%Y-%m-%d %H:%M:%S')

    kwargs['latest_crawl_timestamp'] = time.mktime(latest_crawl_datetime.timetuple())
    kwargs['latest_scrape_timestamp'] = time.mktime(latest_scrape_datetime.timetuple())
    kwargs['latest_matches'] = latest_matches

    # Extract log count and details
    logs_count_tuples = [
        ('critical_logs', 'log_critical_logs_details', 'log_critical_count'),
        ('error_logs', 'log_error_logs_details', 'log_error_count'),
        ('warning_logs', 'log_warning_logs_details', 'log_warning_count'),
        ('redirect_logs', 'redirect_logs_details', 'log_redirect_count'),
        ('retry_logs', 'retry_logs_details', 'log_retry_count'),
        ('ignore_logs', 'ignore_logs_details', 'log_ignore_count'),
    ]
    text += '2018 DEBUG'
    re_matches = []
    for idx, (pattern, (log_title, log_details, log_count)) in enumerate(
            zip(pattern_logs_count_list, logs_count_tuples)):
        matches = pattern.findall(text)
        count = len(matches)
        kwargs[log_count] = count

        re_matches.append({
            'log_title': log_title,
            'log_details': log_details,
            'count': count,
            'line': matches[-1] if matches else '',
            # 'lines': '\n'.join(matches) #if idx < 3 else '',
            'lines': matches
        })
    kwargs['re_matches'] = re_matches

    # 'finish_reason': 'closespider_timeout',
    m = re.search(r":\s'(.*?)'", re_search_final_match(r"'finish_reason'"))
    kwargs['finish_reason'] = m.group(1) if m else ''

    if kwargs['finish_reason']:
        m = re.search(r":\s(\d+)", re_search_final_match(r"'downloader/response_status_count/200'"))
        kwargs['crawled_pages'] = max(kwargs['crawled_pages'], int(m.group(1)) if m else 0)
        m = re.search(r":\s(\d+)", re_search_final_match(r"'item_scraped_count'"))
        kwargs['scraped_items'] = (int(m.group(1)) if m else 0) or kwargs['scraped_items']

        m = re.search(r":\s(\d+)", re_search_final_match(r"'log_count/CRITICAL'"))
        kwargs['log_critical_count'] = int(m.group(1)) if m else kwargs['log_critical_count']
        m = re.search(r":\s(\d+)", re_search_final_match(r"'log_count/ERROR'"))
        kwargs['log_error_count'] = int(m.group(1)) if m else kwargs['log_error_count']
        m = re.search(r":\s(\d+)", re_search_final_match(r"'log_count/WARNING'"))
        kwargs['log_warning_count'] = int(m.group(1)) if m else kwargs['log_warning_count']

        m_301 = re.search(r":\s(\d+)", re_search_final_match(r"'downloader/response_status_count/301'"))
        count_301 = int(m_301.group(1)) if m_301 else 0
        m_302 = re.search(r":\s(\d+)", re_search_final_match(r"'downloader/response_status_count/302'"))
        count_302 = int(m_302.group(1)) if m_302 else 0
        kwargs['log_redirect_count'] = (count_301 + count_302) or kwargs['log_redirect_count']

        m = re.search(r":\s(\d+)", re_search_final_match(r"'retry/count'"))
        kwargs['log_retry_count'] = int(m.group(1)) if m else kwargs['log_retry_count']
        m = re.search(r":\s(\d+)", re_search_final_match(r"'httperror/response_ignored_count'"))
        kwargs['log_ignore_count'] = int(m.group(1)) if m else kwargs['log_ignore_count']



# coding: utf8
from flask import url_for, redirect

from .myview import MyView


class IndexView(MyView):

    def __init__(self):
        super(self.__class__, self).__init__()

    def dispatch_request(self, **kwargs):
        if len(self.SCRAPYD_SERVERS) == 1:
            if self.IS_MOBILE and not self.IS_IPAD:
                return redirect(url_for('dashboard', node=self.node, ui='mobile'))
            else:
                return redirect(url_for('dashboard', node=self.node, ui=self.UI))
        else:
            if self.IS_MOBILEUI or (self.IS_MOBILE and not self.IS_IPAD):
                return redirect(url_for('dashboard', node=self.node, ui='mobile'))
            else:
                return redirect(url_for('overview', node=self.node, ui=self.UI))



# coding: utf8

__title__ = 'scrapydweb'
__version__ = '1.0.0rc2'
__author__ = 'my8100'
__author_email__ = 'my8100@gmail.com'
__url__ = 'https://github.com/my8100/scrapydweb'
__license__ = 'GNU General Public License v3.0'
__description__ = ("A full-featured web UI for Scrapyd cluster management, "
                   "with Scrapy log analysis & visualization supported.")



# coding: utf8
import os
import io
import time
import re
from collections import OrderedDict
import pickle

from flask import Blueprint, render_template, request, url_for, send_from_directory, redirect

from ..myview import MyView
from .utils import slot
from ..vars import SCHEDULE_PATH, DEFAULT_LATEST_VERSION, UA_DICT


HISTORY_LOG = os.path.join(SCHEDULE_PATH, 'history.log')
if not os.path.exists(HISTORY_LOG):
    with io.open(HISTORY_LOG, 'w') as f:
        f.write(u'history.log')


def generate_cmd(auth, url, data):
    if auth:
        cmd = 'curl -u %s:%s %s' % (auth[0], auth[1], url)
    else:
        cmd = 'curl %s' % url

    for key, value in data.items():
        if key == 'setting':
            for v in value:
                if re.match(r'USER_AGENT=', v):
                    cmd += ' --data-urlencode "setting=%s=%s"' % (tuple(v.split('=', 1)))
                else:
                    cmd += ' -d setting=%s=%s' % (tuple(v.split('=', 1)))
        else:
            cmd += ' -d %s=%s' % (key, value)

    return cmd


bp = Blueprint('schedule', __name__, url_prefix='/')


@bp.route('/schedule/<filename>')
def history(filename):
    return send_from_directory(SCHEDULE_PATH, filename, mimetype='text/plain')


class ScheduleView(MyView):

    def __init__(self):
        super(self.__class__, self).__init__()

        self.project = self.view_args['project']
        self.version = self.view_args['version']
        self.spider = self.view_args['spider']

        self.url = 'http://%s/schedule.json' % self.SCRAPYD_SERVER
        self.template = 'scrapydweb/schedule.html'

    def dispatch_request(self, **kwargs):
        if self.POST:
            selected_nodes = self.get_selected_nodes()
            first_selected_node = selected_nodes[0]
        else:
            if self.project:
                # START button of Dashboard page / Run Spider button of Logs page
                selected_nodes = [self.node]
            else:
                selected_nodes = []
            first_selected_node = self.node

        kwargs = dict(
            node=self.node,
            project=self.project,
            version=self.version,
            spider=self.spider,
            # jobid=self.get_now_string(),
            url=self.url,
            selected_nodes=selected_nodes,
            first_selected_node=first_selected_node,
            url_overview=url_for('overview', node=self.node, opt='schedule'),
            url_schedule_run=url_for('schedule.run', node=self.node),
            url_schedule_history=url_for('schedule.history', filename='history.log'),
            url_listprojects=url_for('api', node=self.node, opt='listprojects'),
            url_listversions=url_for('api', node=self.node, opt='listversions', project='PROJECT_PLACEHOLDER'),
            url_listspiders=url_for('api', node=self.node, opt='listspiders', project='PROJECT_PLACEHOLDER',
                                    version_spider_job='VERSION_PLACEHOLDER'),
            url_schedule_check=url_for('schedule.check', node=self.node)
        )

        return render_template(self.template, **kwargs)


class CheckView(MyView):

    def __init__(self):
        super(self.__class__, self).__init__()

        self.url = 'http://%s/schedule.json' % self.SCRAPYD_SERVER
        self.template = 'scrapydweb/schedule.html'

        self.filename = ''
        self.data = OrderedDict()
        self.slot = slot

    def dispatch_request(self, **kwargs):
        self.prepare_data()
        self.logger.debug(self.data)

        cmd = generate_cmd(self.AUTH, self.url, self.data)
        # '-d' may be in project name, like 'ScrapydWeb-demo'
        cmd = re.sub(r'\s+-d\s+', ' \\\r\n-d ', cmd)
        cmd = re.sub(r'\s+--data-urlencode\s+', ' \\\r\n--data-urlencode ', cmd)
        return self.json_dumps({'filename': self.filename, 'cmd': cmd})

    def prepare_data(self):
        for k, d in [('project', 'projectname'), ('_version', DEFAULT_LATEST_VERSION),
                     ('spider', 'spidername')]:
            self.data[k] = request.form.get(k, d)
        if self.data['_version'] == DEFAULT_LATEST_VERSION:
            self.data.pop('_version')

        self.data['jobid'] = re.sub(r'[^0-9A-Za-z_-]', '', request.form.get('jobid', '')) or self.get_now_string()

        self.data['setting'] = []
        ua = UA_DICT.get(request.form.get('USER_AGENT', ''), '')
        if ua:
            self.data['setting'].append('USER_AGENT=%s' % ua)

        for key in ['COOKIES_ENABLED', 'ROBOTSTXT_OBEY', 'CONCURRENT_REQUESTS', 'DOWNLOAD_DELAY']:
            value = request.form.get(key, '')
            if value:
                self.data['setting'].append("%s=%s" % (key, value))

        additional = request.form.get('additional', '').strip()
        if additional:
            parts = [i.strip() for i in re.split(r'-d\s+', re.sub(r'[\r\n]', ' ', additional)) if i.strip()]
            for part in parts:
                part = re.sub(r'\s*=\s*', '=', part)
                if '=' not in part:
                    continue
                m_setting = re.match(r'setting=([A-Z_]{6,31}=.+)', part)  # 'EDITOR' 'DOWNLOADER_CLIENTCONTEXTFACTORY'
                if m_setting:
                    self.data['setting'].append(m_setting.group(1))
                    continue
                m_arg = re.match(r'([a-zA-Z_][0-9a-zA-Z_]*)=(.+)', part)
                if m_arg and m_arg.group(1) != 'setting':
                    self.data[m_arg.group(1)] = m_arg.group(2)

        self.data['setting'].sort()
        _version = self.data.get('_version', 'default-the-latest-version')
        _filename = '{project}_{version}_{spider}'.format(project=self.data['project'],
                                                          version=_version,
                                                          spider=self.data['spider'])
        self.filename = '%s.pickle' % re.sub(r'[^0-9A-Za-z_-]', '', _filename)
        filepath = os.path.join(SCHEDULE_PATH, self.filename)
        with io.open(filepath, 'wb') as f:
            f.write(pickle.dumps(self.data))

        self.slot.add_data(self.filename, self.data)


class RunView(MyView):

    def __init__(self):
        super(self.__class__, self).__init__()

        self.url = ''
        self.template = 'scrapydweb/schedule_results.html'

        self.slot = slot
        self.selected_nodes_amount = 0
        self.selected_nodes = []
        self.first_selected_node = 0
        self.filename = request.form['filename']
        self.data = None
        self.js = {}

    def dispatch_request(self, **kwargs):
        self.handle_form()
        status_code, self.js = self.make_request(self.url, self.data, auth=self.AUTH)
        self.update_history()
        return self.generate_response()

    def handle_form(self):
        self.selected_nodes_amount = int(request.form.get('checked_amount', 0))
        # With multinodes, would try to Schedule to the first selected node first
        if self.selected_nodes_amount:
            self.selected_nodes = self.get_selected_nodes()
            self.first_selected_node = self.selected_nodes[0]
            self.url = 'http://%s/schedule.json' % self.SCRAPYD_SERVERS[self.first_selected_node - 1]
            # Note that self.first_selected_node != self.node
            self.AUTH = self.SCRAPYD_SERVERS_AUTHS[self.first_selected_node - 1]
        else:
            self.selected_nodes = [self.node]
            self.url = 'http://%s/schedule.json' % self.SCRAPYD_SERVER

        self.data = self.slot.data.get(self.filename)
        # self.data = None  # For test only
        if not self.data:
            filepath = os.path.join(SCHEDULE_PATH, self.filename)
            with io.open(filepath, 'rb') as f:
                self.data = pickle.loads(f.read())

    def update_history(self):
        with io.open(HISTORY_LOG, 'r+', encoding='utf8') as f:
            content_old = f.read()
            f.seek(0)
            content = os.linesep.join([
                '#' * 50,
                time.ctime(),
                str([self.SCRAPYD_SERVERS[i - 1] for i in self.selected_nodes]),
                generate_cmd(self.AUTH, self.url, self.data),
                self.json_dumps(self.js),
                ''
            ])
            f.write(content)
            f.write(content_old)

    def generate_response(self):
        if self.js['status'] == 'ok':
            if not self.selected_nodes_amount:
                return redirect(url_for('dashboard', node=self.node))

            kwargs = dict(
                node=self.node,
                project=self.data['project'],
                version=self.data.get('_version', DEFAULT_LATEST_VERSION),
                spider=self.data['spider'],
                selected_nodes=self.selected_nodes,
                first_selected_node=self.first_selected_node,
                js=self.js,
                url_dashboard_first_selected_node=url_for('dashboard', node=self.first_selected_node),
                url_xhr=url_for('schedule.schedule_xhr', node=self.node, filename=self.filename),
                url_overview=url_for('overview', node=self.node, opt='stop', project=self.data['project'],
                                     version_job=self.js['jobid'])
            )
            return render_template(self.template, **kwargs)
        else:
            message = self.js.get('message', '')
            if message:
                self.js.update({'message': 'See details below'})
            self.js['info'] = "Maybe the project egg file had been deleted, check out the Manage page."

            if self.selected_nodes_amount > 1:
                alert = ("Multinode schedule terminated, "
                         "since the first selected node returned status: " + self.js['status'])
            else:
                alert = "Fail to schedule, got status: " + self.js['status']

            return render_template(self.template_fail, node=self.node,
                                   alert=alert, text=self.json_dumps(self.js), message=message)


class ScheduleXhrView(MyView):

    def __init__(self):
        super(self.__class__, self).__init__()

        self.filename = self.view_args['filename']

        self.url = 'http://%s/schedule.json' % self.SCRAPYD_SERVER

        self.slot = slot
        self.data = None

    def dispatch_request(self, **kwargs):
        self.data = self.slot.data.get(self.filename)
        # self.data = None  # For test only
        if not self.data:
            filepath = os.path.join(SCHEDULE_PATH, self.filename)
            with io.open(filepath, 'rb') as f:
                self.data = pickle.loads(f.read())

        status_code, js = self.make_request(self.url, self.data, auth=self.AUTH)
        return self.json_dumps(js)



# coding: utf8
import os
import sys
from shutil import rmtree
import re

from ..vars import ALLOWED_SCRAPYD_LOG_EXTENSIONS, CACHE_PATH, EMAIL_TRIGGER_KEYS
from .utils import printf, json_dumps
from .send_email import send_email


pattern_email = re.compile(r'^[^@]+@[^@]+\.[^@]+$')


def check_app_config(config):
    def check_assert(key, default, is_instance, allow_zero=True, non_empty=False, containing_type=None):
        if is_instance is int:
            if allow_zero:
                should_be = "a non-negative integer"
            else:
                should_be = "a positive integer"
        else:
            should_be = "an instance of %s%s" % (is_instance, ' and NOT empty' if non_empty else '')

        value = config.get(key, default)
        kwargs = dict(
            key=key,
            should_be=should_be,
            containing_type=', containing elements of type %s' % containing_type if containing_type else '',
            value="'%s'" % value if isinstance(value, str) else value
        )
        to_assert = "{key} should be {should_be}{containing_type}. Current value: {value}".format(**kwargs)

        assert (isinstance(value, is_instance)
                and (not isinstance(value, bool) if is_instance is int else True)  # isinstance(True, int) => True
                and (value > (-1 if allow_zero else 0) if is_instance is int else True)
                and (value if non_empty else True)
                and (all([isinstance(i, containing_type) for i in value]) if containing_type else True)), to_assert
    printf("Checking app config:")

    # ScrapydWeb
    check_assert('SCRAPYDWEB_BIND', '0.0.0.0', str, non_empty=True)
    SCRAPYDWEB_PORT = config.get('SCRAPYDWEB_PORT', 5000)
    try:
        assert not isinstance(SCRAPYDWEB_PORT, bool)
        SCRAPYDWEB_PORT = int(SCRAPYDWEB_PORT)
        assert SCRAPYDWEB_PORT > 0
    except (ValueError, AssertionError):
        sys.exit("SCRAPYDWEB_PORT should be a positive integer. Current value: %s" % SCRAPYDWEB_PORT)

    check_assert('ENABLE_AUTH', False, bool)
    if config.get('ENABLE_AUTH', False):
        # May be 0 from config file
        check_assert('USERNAME', '', str, non_empty=True)
        check_assert('PASSWORD', '', str, non_empty=True)
        printf("Basic auth enabled with USERNAME/PASSWORD: '%s'/'%s'" % (config['USERNAME'], config['PASSWORD']))

    # Scrapy
    check_assert('SCRAPY_PROJECTS_DIR', '', str)
    SCRAPY_PROJECTS_DIR = config.get('SCRAPY_PROJECTS_DIR', '')
    if SCRAPY_PROJECTS_DIR:
        if not os.path.isdir(SCRAPY_PROJECTS_DIR):
            sys.exit("!!! SCRAPY_PROJECTS_DIR NOT found: %s" % SCRAPY_PROJECTS_DIR)
        else:
            printf("Using SCRAPY_PROJECTS_DIR: %s" % SCRAPY_PROJECTS_DIR)

    # Scrapyd
    # print(config.get('SCRAPYD_SERVERS'))  # Checked in the preceding update_app_config()
    check_assert('SCRAPYD_LOGS_DIR', '', str)
    SCRAPYD_LOGS_DIR = config.get('SCRAPYD_LOGS_DIR', '')
    if SCRAPYD_LOGS_DIR:
        if not os.path.isdir(SCRAPYD_LOGS_DIR):
            sys.exit("!!! SCRAPYD_LOGS_DIR NOT found: %s" % SCRAPYD_LOGS_DIR)
        else:
            printf("Using SCRAPYD_LOGS_DIR: %s" % SCRAPYD_LOGS_DIR)

    check_assert('SCRAPYD_LOG_EXTENSIONS', ALLOWED_SCRAPYD_LOG_EXTENSIONS, list, non_empty=True, containing_type=str)
    SCRAPYD_LOG_EXTENSIONS = config.get('SCRAPYD_LOG_EXTENSIONS', ALLOWED_SCRAPYD_LOG_EXTENSIONS)
    assert all([not i or i.startswith('.') for i in SCRAPYD_LOG_EXTENSIONS]), \
        ("SCRAPYD_LOG_EXTENSIONS should be a list like %s. "
         "Current value: %s" % (ALLOWED_SCRAPYD_LOG_EXTENSIONS, SCRAPYD_LOG_EXTENSIONS))
    printf("Locating scrapy log with SCRAPYD_LOG_EXTENSIONS: %s" % SCRAPYD_LOG_EXTENSIONS)

    # Page Display
    check_assert('SHOW_SCRAPYD_ITEMS', True, bool)
    check_assert('SHOW_DASHBOARD_JOB_COLUMN', False, bool)
    check_assert('DASHBOARD_RELOAD_INTERVAL', 300, int)
    check_assert('DAEMONSTATUS_REFRESH_INTERVAL', 10, int)

    # HTML Caching
    check_assert('ENABLE_CACHE', True, bool)
    if config.get('ENABLE_CACHE', True):
        check_assert('CACHE_ROUND_INTERVAL', 300, int, allow_zero=False)
        check_assert('CACHE_REQUEST_INTERVAL', 10, int, allow_zero=False)

    check_assert('DELETE_CACHE', False, bool)
    if config.get('DELETE_CACHE', False):
        if os.path.isdir(CACHE_PATH):
            rmtree(CACHE_PATH, ignore_errors=True)
            printf("Cached HTML files of Log and Stats page deleted")
        else:
            printf("Cache dir NOT found: %s" % CACHE_PATH, warn=True)

    # Email Notice
    check_assert('ENABLE_EMAIL', False, bool)
    if config.get('ENABLE_EMAIL', False):
        check_assert('SMTP_SERVER', '', str, non_empty=True)
        check_assert('SMTP_PORT', 0, int, allow_zero=False)
        check_assert('SMTP_OVER_SSL', False, bool)
        check_assert('SMTP_CONNECTION_TIMEOUT', 10, int, allow_zero=False)

        check_assert('FROM_ADDR', '', str, non_empty=True)
        FROM_ADDR = config['FROM_ADDR']
        assert pattern_email.search(FROM_ADDR), \
            "FROM_ADDR should contain '@', like 'username@gmail.com'. Current value: %s" % FROM_ADDR
        check_assert('EMAIL_PASSWORD', '', str, non_empty=True)
        check_assert('TO_ADDRS', [], list, non_empty=True, containing_type=str)
        TO_ADDRS = config['TO_ADDRS']
        assert all([pattern_email.search(i) for i in TO_ADDRS]), \
            "All elements in TO_ADDRS should contain '@', like 'username@gmail.com'. Current value: %s" % TO_ADDRS

        # For compatibility with Python 3 using range()
        try:
            config['EMAIL_WORKING_DAYS'] = list(config.get('EMAIL_WORKING_DAYS', []))
        except TypeError:
            pass
        check_assert('EMAIL_WORKING_DAYS', [], list, non_empty=True, containing_type=int)
        EMAIL_WORKING_DAYS = config['EMAIL_WORKING_DAYS']
        assert all([not isinstance(i, bool) and i in range(1, 8) for i in EMAIL_WORKING_DAYS]), \
            "Element in EMAIL_WORKING_DAYS should be between 1 and 7. Current value: %s" % EMAIL_WORKING_DAYS

        try:
            config['EMAIL_WORKING_HOURS'] = list(config.get('EMAIL_WORKING_HOURS', []))
        except TypeError:
            pass
        check_assert('EMAIL_WORKING_HOURS', [], list, non_empty=True, containing_type=int)
        EMAIL_WORKING_HOURS = config['EMAIL_WORKING_HOURS']
        assert all([not isinstance(i, bool) and i in range(24) for i in EMAIL_WORKING_HOURS]), \
            "Element in EMAIL_WORKING_HOURS should be between 0 and 23. Current value: %s" % EMAIL_WORKING_HOURS

        check_assert('ON_JOB_RUNNING_INTERVAL', 0, int)
        check_assert('ON_JOB_FINISHED', False, bool)

        for k in EMAIL_TRIGGER_KEYS:
            check_assert('LOG_%s_THRESHOLD' % k, 0, int)
            check_assert('LOG_%s_TRIGGER_STOP' % k, False, bool)
            check_assert('LOG_%s_TRIGGER_FORCESTOP' % k, False, bool)

        check_email(config)

    # System
    check_assert('DEBUG', False, bool)
    if config.get('DEBUG', False):
        printf("It's not recommended to run ScrapydWeb in debug mode, set 'DEBUG = False' instead", warn=True)
    check_assert('VERBOSE', False, bool)


def check_email(config):
    assert config.get('ENABLE_CACHE', True), \
        ("In order to enable 'Email Notice', you have to enable 'HTML Caching' by setting "
         "'ENABLE_CACHE = True' first,\nalso, don't pass in the argument '--disable_cache'")

    kwargs = dict(
        smtp_server=config['SMTP_SERVER'],
        smtp_port=config['SMTP_PORT'],
        smtp_over_ssl=config.get('SMTP_OVER_SSL', False),
        smtp_connection_timeout=config.get('SMTP_CONNECTION_TIMEOUT', 10),
        from_addr=config['FROM_ADDR'],
        email_password=config['EMAIL_PASSWORD'],
        to_addrs=config['TO_ADDRS']
    )
    kwargs['to_retry'] = True
    kwargs['subject'] = 'Email notice enabled #scrapydweb'
    kwargs['content'] = json_dumps(dict(FROM_ADDR=config['FROM_ADDR'], TO_ADDRS=config['TO_ADDRS']))

    printf("Trying to send email (smtp_connection_timeout=%s)..." % config.get('SMTP_CONNECTION_TIMEOUT', 10))
    result = send_email(**kwargs)
    if not result:
        print('')
        print(json_dumps(kwargs, sort_keys=False))
    assert result, "Fail to send email. Modify the email settings above or pass in the argument '--disable_email'"

    printf("Email notice enabled")



# -*- coding: utf-8 -*-
import scrapy


class TestSpider(scrapy.Spider):
    name = 'test'
    allowed_domains = ['httpbin.org', 'google.com']
    # start_urls = ['http://httpbin.org/']

    custom_settings = {
        # 'JOBDIR': 'crawls/test',
        'ROBOTSTXT_OBEY': False,
        'USER_AGENT': 'Mozilla/5.0',
        'COOKIES_ENABLED': False,
        'CONCURRENT_REQUESTS': 2,
        'DOWNLOAD_DELAY': 1,
        'LOGSTATS_INTERVAL': 1,
    }

    num = 1

    def start_requests(self):
        self.log(u'test unicode: 测试中文')
        self.logger.debug('2018-08-20 09:13:06 [apps_redis] DEBUG: Resuming crawl (675840 requests scheduled)')

        self.logger.warning('warn') #python2.7 AttributeError: 'LoggerAdapter' object has no attribute 'warn'
        self.logger.error('error')
        self.logger.warning('warning\n123abc')
        self.logger.error('error\n456abc')
        self.logger.error('error\n456abc')
        self.logger.critical('critical\n789abc')
        self.logger.warning('warning\n 123\nabc')
        self.logger.error('error\n 456\nabc')
        self.logger.critical('critical')
        self.logger.critical('critical\n 789\nabc')
        self.logger.critical('critical')
        self.logger.critical('critical')

        yield scrapy.Request('http://httpbin.org/redirect/1')
        yield scrapy.Request('http://httpbin.org/status/404')
        yield scrapy.Request('http://httpbin.org/headers')
        yield scrapy.Request('http://httpbin.org/headers')
        yield scrapy.Request('https://google.com/')

    def parse(self, response):
        if self.num == 1:
            yield scrapy.Request('https://www.baidu.com/')

        yield {'item': self.num}

        self.num += 1



# coding: utf8
import os
import json

from flask import Response


def printf(value, warn=False):
    prefix = "!!!" if warn else ">>>"
    print("%s %s" % (prefix, value))


def json_dumps(obj, sort_keys=True):
    return json.dumps(obj, ensure_ascii=False, indent=4, sort_keys=sort_keys)


def find_scrapydweb_settings_py(filename, path, prevpath=None):
    if path == prevpath:
        return ''
    path = os.path.abspath(path)
    cfgfile = os.path.join(path, filename)
    if os.path.exists(cfgfile):
        return cfgfile
    return find_scrapydweb_settings_py(filename, os.path.dirname(path), path)


# http://flask.pocoo.org/snippets/category/authentication/
def authenticate():
    """Sends a 401 response that enables basic auth"""
    return Response("<script>alert('FAIL to login: basic auth for ScrapydWeb has been enabled');</script>",
                    401, {'WWW-Authenticate': 'Basic realm="ScrapydWeb Basic Auth Required"'})



# coding: utf8
import os
import sys
import io
import glob
import time
from datetime import datetime
import re
import tempfile
import zipfile
import tarfile
from shutil import copyfile, rmtree, copyfileobj

from .scrapyd_deploy import _build_egg
from flask import render_template, request, url_for, redirect
from werkzeug.utils import secure_filename

from ..myview import MyView
from .utils import slot, mkdir_p
from ..vars import DEPLOY_PATH


PY2 = True if sys.version_info[0] < 3 else False


class DeployView(MyView):

    def __init__(self):
        super(self.__class__, self).__init__()

        self.url = 'http://{}/{}.json'.format(self.SCRAPYD_SERVER, 'addversion')
        self.template = 'scrapydweb/deploy.html'

    def dispatch_request(self, **kwargs):
        # python2 'ascii' codec can't decode byte
        scrapy_cfg_list = glob.glob(os.path.join(self.SCRAPY_PROJECTS_DIR, '*', u'scrapy.cfg'))
        projects_list = [os.path.dirname(i) for i in scrapy_cfg_list]

        kwargs = dict(
            node=self.node,
            url=self.url,
            selected_nodes=self.get_selected_nodes(),
            projects=[os.path.basename(i) for i in projects_list],
            modification_times=[self.get_modification_time(i) for i in projects_list],
            SCRAPY_PROJECTS_DIR=self.SCRAPY_PROJECTS_DIR,
            url_overview=url_for('overview', node=self.node, opt='deploy'),
            url_deploy_upload=url_for('deploy.upload', node=self.node)
        )
        return render_template(self.template, **kwargs)

    @staticmethod
    def get_modification_time(path):
        # https://stackoverflow.com/a/29685234/10517783
        # https://stackoverflow.com/a/13454267/10517783
        filepath_list = []

        in_top_dir = True
        for dirpath, dirnames, filenames in os.walk(path):
            if in_top_dir:
                in_top_dir = False
                dirnames[:] = [d for d in dirnames if d not in ['build', 'project.egg-info']]
                filenames = [f for f in filenames if not f.endswith('.egg')]
            for filename in filenames:
                filepath_list.append(os.path.join(dirpath, filename))
        max_timestamp = max([os.path.getmtime(f) for f in filepath_list] or [time.time()])
        return datetime.fromtimestamp(max_timestamp).strftime('%Y-%m-%dT%H_%M_%S')


class UploadView(MyView):
    methods = ['POST']

    def __init__(self):
        super(self.__class__, self).__init__()

        self.url = ''
        self.template = 'scrapydweb/deploy_results.html'

        self.project_original = ''
        self.project = ''
        self.version = ''
        self.selected_nodes_amount = 0
        self.selected_nodes = []
        self.first_selected_node = 0

        self.eggname = ''
        self.eggpath = ''
        self.scrapy_cfg_path = ''
        self.scrapy_cfg_searched_paths = []
        self.scrapy_cfg_not_found = False
        self.data = None

        self.slot = slot

    def dispatch_request(self, **kwargs):
        self.handle_form()

        if self.scrapy_cfg_not_found:
            text = "scrapy.cfg NOT found"
            if self.selected_nodes_amount > 1:
                alert = "Multinode deployment terminated: %s" % text
            else:
                alert = "Fail to deploy project: %s" % text
            return render_template(self.template_fail, node=self.node,
                                   alert=alert, text=text,
                                   message=self.json_dumps(self.scrapy_cfg_searched_paths))
        else:
            self.prepare_data()
            status_code, js = self.make_request(self.url, self.data, auth=self.AUTH)

        if js['status'] != 'ok':
            # With multinodes, would try to deploy to the first selected node first
            if self.selected_nodes_amount > 1:
                alert = ("Multinode deployment terminated, "
                         "since the first selected node returned status: " + js['status'])
            else:
                alert = "Fail to deploy project, got status: " + js['status']
            message = js.get('message', '')
            if message:
                js.update({'message': 'See details below'})

            return render_template(self.template_fail, node=self.node,
                                   alert=alert, text=self.json_dumps(js), message=message)
        else:
            if self.selected_nodes_amount == 0:
                return redirect(url_for('schedule.schedule', node=self.node,
                                        project=self.project, version=self.version))
            else:
                kwargs = dict(
                    node=self.node,
                    selected_nodes=self.selected_nodes,
                    first_selected_node=self.first_selected_node,
                    js=js,
                    project=self.project,
                    version=self.version,
                    url_manage_first_selected_node=url_for('manage', node=self.first_selected_node),
                    url_manage_list=[url_for('manage', node=n) for n in range(1, len(self.SCRAPYD_SERVERS)+1)],
                    url_xhr=url_for('deploy.deploy_xhr', node=self.node, eggname=self.eggname,
                                    project=self.project, version=self.version),
                    url_schedule=url_for('schedule.schedule', node=self.node, project=self.project,
                                         version=self.version),
                    url_overview=url_for('overview', node=self.node, opt='schedule', project=self.project,
                                         version_job=self.version)
                )
                return render_template(self.template, **kwargs)

    def handle_form(self):
        # {'1': 'on',
        # '2': 'on',
        # 'checked_amount': '2',
        # 'project': 'demo',
        # 'version': '2018-09-05T03_13_50'}

        # With multinodes, would try to deploy to the first selected node first
        self.selected_nodes_amount = int(request.form.get('checked_amount', 0))
        if self.selected_nodes_amount:
            self.selected_nodes = self.get_selected_nodes()
            self.first_selected_node = self.selected_nodes[0]
            self.url = 'http://{}/{}.json'.format(self.SCRAPYD_SERVERS[self.first_selected_node - 1], 'addversion')
            # Note that self.first_selected_node != self.node
            self.AUTH = self.SCRAPYD_SERVERS_AUTHS[self.first_selected_node - 1]
        else:
            self.url = 'http://{}/{}.json'.format(self.SCRAPYD_SERVER, 'addversion')

        self.project_original = request.form.get('project', '')  # Used with SCRAPY_PROJECTS_DIR to get project_path
        self.project = re.sub(r'[^0-9A-Za-z_-]', '', self.project_original) or self.get_now_string()
        self.version = re.sub(r'[^0-9A-Za-z_-]', '', request.form.get('version', '')) or self.get_now_string()

        if request.files.get('file'):
            self.handle_uploaded_file()
        else:
            self.handle_local_project()

    def handle_local_project(self):
        # Use project_original instead of project
        project_path = os.path.join(self.SCRAPY_PROJECTS_DIR, self.project_original)

        self.search_scrapy_cfg_path(project_path)
        if not self.scrapy_cfg_path:
            self.scrapy_cfg_not_found = True
            return

        self.eggname = '%s_%s.egg' % (self.project, self.version)
        self.eggpath = os.path.join(DEPLOY_PATH, self.eggname)
        self.build_egg()

    def handle_uploaded_file(self):
        # http://flask.pocoo.org/docs/1.0/api/#flask.Request.form
        # <class 'werkzeug.datastructures.FileStorage'>
        file = request.files['file']

        # Non-ASCII would be omitted and resulting the filename as to 'egg' or 'tar.gz'
        filename = secure_filename(file.filename)
        # tar.xz only works on Linux and macOS
        if filename in ['egg', 'zip', 'tar.gz']:
            filename = '%s_%s.%s' % (self.project, self.version, filename)
        else:
            filename = '%s_%s_from_file_%s' % (self.project, self.version, filename)

        if filename.endswith('egg'):
            self.eggname = filename
            self.eggpath = os.path.join(DEPLOY_PATH, self.eggname)
            file.save(self.eggpath)
            self.scrapy_cfg_not_found = False
        else:  # Compressed file
            filepath = os.path.join(DEPLOY_PATH, filename)
            file.save(filepath)
            tmpdir = self.uncompress_to_tmpdir(filepath)

            # Search from the root of tmpdir
            self.search_scrapy_cfg_path(tmpdir)
            if not self.scrapy_cfg_path:
                self.scrapy_cfg_not_found = True
                return

            self.eggname = re.sub(r'(\.zip|\.tar\.gz)$', '.egg', filename)
            self.eggpath = os.path.join(DEPLOY_PATH, self.eggname)
            self.build_egg()

    # https://gangmax.me/blog/2011/09/17/12-14-52-publish-532/
    # https://stackoverflow.com/a/49649784
    # When ScrapydWeb runs in Linux/macOS and tries to uncompress zip file from Windows_CN_cp936
    # UnicodeEncodeError: 'ascii' codec can't encode characters in position 7-8: ordinal not in range(128)
    # macOS + PY2 would raise OSError: Illegal byte sequence
    # Ubuntu + PY2 would raise UnicodeDecodeError in search_scrapy_cfg_path() though f.extractall(tmpdir) works well
    def uncompress_to_tmpdir(self, filepath):
        self.logger.debug("Uncompress %s" % filepath)
        tmpdir = tempfile.mkdtemp(prefix="scrapydweb-uncompress-")
        if zipfile.is_zipfile(filepath):
            with zipfile.ZipFile(filepath, 'r') as f:
                if PY2:
                    tmpdir = tempfile.mkdtemp(prefix="scrapydweb-uncompress-")
                    for filename in f.namelist():
                        try:
                            filename_utf8 = filename.decode('gbk').encode('utf8')
                        except (UnicodeDecodeError, UnicodeEncodeError):
                            filename_utf8 = filename
                        filepath_utf8 = os.path.join(tmpdir, filename_utf8)

                        try:
                            with io.open(filepath_utf8, 'wb') as f_utf8:
                                copyfileobj(f.open(filename), f_utf8)
                        except IOError:
                            # os.mkdir(filepath_utf8)
                            # zipfile from Windows "send to zipped" would meet the inner folder first:
                            # temp\\scrapydweb-uncompress-qrcyc0\\demo7/demo/'
                            mkdir_p(filepath_utf8)
                else:
                    f.extractall(tmpdir)
        else:  # tar.gz
            with tarfile.open(filepath, 'r') as f:  # Open for reading with transparent compression (recommended).
                f.extractall(tmpdir)
                f.close()

        self.logger.debug("Uncompress to %s" % tmpdir)
        # In case uploading a compressed file in which scrapy_cfg_dir contains none ascii in python 2,
        # whereas selecting a project when auto eggifying, scrapy_cfg_dir is unicode
        # print(repr(tmpdir))
        # print(type(tmpdir))
        return tmpdir.decode('utf8') if PY2 else tmpdir

    def search_scrapy_cfg_path(self, search_path):
        for dirpath, dirnames, filenames in os.walk(search_path):
            self.scrapy_cfg_searched_paths.append(os.path.abspath(dirpath))
            self.scrapy_cfg_path = os.path.abspath(os.path.join(dirpath, 'scrapy.cfg'))
            if os.path.exists(self.scrapy_cfg_path):
                self.logger.debug("scrapy_cfg_path: %s" % self.scrapy_cfg_path)
                return

        self.logger.error("scrapy.cfg NOT found in: %s" % search_path)
        self.scrapy_cfg_path = ''

    def build_egg(self):
        egg, tmpdir = _build_egg(self.scrapy_cfg_path)

        scrapy_cfg_dir = os.path.dirname(self.scrapy_cfg_path)
        copyfile(egg, os.path.join(scrapy_cfg_dir, self.eggname))
        copyfile(egg, self.eggpath)
        rmtree(tmpdir)
        self.logger.debug("egg file saved to: %s" % self.eggpath)

    def prepare_data(self):
        with io.open(self.eggpath, 'rb') as f:
            content = f.read()
            self.data = {
                'project': self.project,
                'version': self.version,
                'egg': content
            }

        self.slot.add_egg(self.eggname, content)


class DeployXhrView(MyView):

    def __init__(self):
        super(self.__class__, self).__init__()

        self.eggname = self.view_args['eggname']
        self.project = self.view_args['project']
        self.version = self.view_args['version']

        self.url = 'http://{}/{}.json'.format(self.SCRAPYD_SERVER, 'addversion')

        self.slot = slot

    def dispatch_request(self, **kwargs):
        content = self.slot.egg.get(self.eggname)
        # content = None  # For test only
        if not content:
            eggpath = os.path.join(DEPLOY_PATH, self.eggname)
            with io.open(eggpath, 'rb') as f:
                content = f.read()

        data = {
            'project': self.project,
            'version': self.version,
            'egg': content
        }
        status_code, js = self.make_request(self.url, data, auth=self.AUTH)
        return self.json_dumps(js)



# coding: utf8
import sys
import time
import json
import smtplib
from email.mime.text import MIMEText


def printf(value, warn=False):
    prefix = "!!!" if warn else ">>>"
    print("%s %s" % (prefix, value))


# https://stackoverflow.com/a/27515833/10517783 How to send an email with Gmail as provider using Python?
# https://stackoverflow.com/a/26053352/10517783 Python smtplib proxy support
def send_email(**kwargs):
    to_retry = kwargs.get('to_retry', False)
    need_debug = kwargs.get('need_debug', False)

    smtp_server = kwargs['smtp_server']
    smtp_port = kwargs['smtp_port']
    smtp_over_ssl = kwargs['smtp_over_ssl']
    smtp_connection_timeout = kwargs['smtp_connection_timeout']
    from_addr = kwargs['from_addr']
    email_password = kwargs['email_password']
    to_addrs = kwargs['to_addrs']
    subject = kwargs['subject']
    content = kwargs['content']

    msg = MIMEText(u'%s\n%s' % (time.ctime(), content), 'plain', 'utf-8')
    msg['From'] = from_addr
    msg['Subject'] = u'{} {}'.format(time.strftime('%H:%M'), subject)

    result = False
    try:
        if smtp_over_ssl:
            server = smtplib.SMTP_SSL(smtp_server, smtp_port, timeout=smtp_connection_timeout)
        else:
            server = smtplib.SMTP(smtp_server, smtp_port, timeout=smtp_connection_timeout)
            server.ehlo()
            server.starttls()
        if need_debug:
            server.set_debuglevel(1)  # For debug
        server.login(from_addr, email_password)
        server.sendmail(from_addr, to_addrs, msg.as_string())
    except Exception as err:
        printf("FAIL to send email: %s" % subject)
        try:
            printf("FAIL reason: %s" % err.args[-1].decode('utf8'))
        except:
            try:
                printf("FAIL reason: %s" % err.args[-1].decode('gbk'))
            except:
                printf("FAIL reason: %s" % err)
        if to_retry:
            kwargs.update({'to_retry': False, 'need_debug': True})
            printf("Retrying...")
            send_email(**kwargs)
    else:
        result = True
        printf("Email sent: %s" % subject)
    finally:
        try:
            server.quit()
        except:
            pass

    return result


if __name__ == '__main__':
    send_email(**json.loads(sys.argv[1]))



# coding: utf8
import re
import traceback
import logging
from logging.config import dictConfig

from flask import Flask
from flask_compress import Compress


title_error_500 = """
<h1>Internal Server Error</h1>
<p>The server encountered an internal error and was unable to complete your request.
Either the server is overloaded or there is an error in the application.</p>
"""

# http://flask.pocoo.org/docs/1.0/logging/#basic-configuration
dictConfig({
    'version': 1,
    'formatters': {'default': {
        'format': '[%(asctime)s] %(levelname)s in %(name)s: %(message)s',
    }},
    'handlers': {'wsgi': {
        'class': 'logging.StreamHandler',
        'stream': 'ext://flask.logging.wsgi_errors_stream',
        'formatter': 'default'
    }},
    'root': {
        'level': 'DEBUG',
        'handlers': ['wsgi']
    }
})


def create_app(test_config=None):
    app = Flask(__name__, instance_relative_config=True)
    app.config.from_mapping(
        SECRET_KEY='dev',
    )
    # http://flask.pocoo.org/docs/1.0/config/#configuring-from-files
    app.config.from_object('scrapydweb.default_settings')

    if test_config is None:
        # load the instance config, if it exists, when not testing
        app.config.from_pyfile('config.py', silent=True)
    else:
        # load the test config if passed in
        app.config.from_mapping(test_config)

    @app.route('/hello')
    def hello():
        return 'Hello, World!'

    # @app.errorhandler(404)
    # def handle_error(error):
        # return ('Nothing Found', 404)

    @app.errorhandler(500)
    def handle_error(error):
        return '{}<h2>error: {}</h2><pre>{}<pre>'.format(title_error_500, error, traceback.format_exc()), 500

    handle_route(app)

    compress = Compress()
    compress.init_app(app)

    app.jinja_env.variable_start_string = '{{ '
    app.jinja_env.variable_end_string = ' }}'

    # https://ansible-docs.readthedocs.io/zh/stable-2.0/rst/playbooks_filters.html#other-useful-filters
    # https://stackoverflow.com/questions/12791216/how-do-i-use-regular-expressions-in-jinja2
    # https://www.michaelcho.me/article/custom-jinja-template-filters-in-flask
    # http://flask.pocoo.org/docs/1.0/api/#flask.Flask.template_filter
    @app.template_filter()
    def regex_replace(s, find, replace):
        return re.sub(find, replace, s)

    app.logger.setLevel(logging.DEBUG)

    return app


def handle_route(app):
    def register_view(view, endpoint, url_defaults_list):
        view_func = view.as_view(endpoint)
        for url, defaults in url_defaults_list:
            app.add_url_rule('/<int:node>/%s/' % url, defaults=defaults, view_func=view_func)

    from .index import IndexView
    index_view = IndexView.as_view('index')
    app.add_url_rule('/<int:node>/', view_func=index_view)
    app.add_url_rule('/', defaults=dict(node=1), view_func=index_view)

    from .api import ApiView
    register_view(ApiView, 'api', [
        ('api/<opt>/<project>/<version_spider_job>', None),
        ('api/<opt>/<project>', dict(version_spider_job=None)),
        ('api/<opt>', dict(project=None, version_spider_job=None))
    ])

    # jobs
    from .jobs.dashboard import DashboardView
    register_view(DashboardView, 'dashboard', [('dashboard', None)])

    from .jobs.overview import OverviewView
    register_view(OverviewView, 'overview', [
        ('overview/<opt>/<project>/<version_job>/<spider>', None),
        ('overview/<opt>/<project>/<version_job>', dict(spider=None)),
        ('overview/<opt>/<project>', dict(version_job=None, spider=None)),
        ('overview/<opt>', dict(project=None, version_job=None, spider=None)),
        ('overview', dict(opt=None, project=None, version_job=None, spider=None))
    ])

    from .jobs.multinode import MultinodeView
    register_view(MultinodeView, 'multinode', [
        ('multinode/<opt>/<project>/<version_job>', None),
        ('multinode/<opt>/<project>', dict(version_job=None))
    ])

    # projects
    from .projects.deploy import DeployView, UploadView, DeployXhrView
    register_view(DeployView, 'deploy.deploy', [('deploy', None)])
    register_view(UploadView, 'deploy.upload', [('deploy/upload', None)])
    register_view(DeployXhrView, 'deploy.deploy_xhr', [('deploy/xhr/<eggname>/<project>/<version>', None)])

    from .projects.schedule import ScheduleView, CheckView, RunView, ScheduleXhrView
    register_view(ScheduleView, 'schedule.schedule', [
        ('schedule/<project>/<version>/<spider>', None),
        ('schedule/<project>/<version>', dict(spider=None)),
        ('schedule/<project>', dict(version=None, spider=None)),
        ('schedule', dict(project=None, version=None, spider=None))
    ])
    register_view(CheckView, 'schedule.check', [('schedule/check', None)])
    register_view(RunView, 'schedule.run', [('schedule/run', None)])
    register_view(ScheduleXhrView, 'schedule.schedule_xhr', [('schedule/xhr/<filename>', None)])

    from .projects.schedule import bp as bp_schedule_history
    app.register_blueprint(bp_schedule_history)

    from .projects.manage import ManageView
    register_view(ManageView, 'manage', [
        ('manage/<opt>/<project>/<version_spider_job>', None),
        ('manage/<opt>/<project>', dict(version_spider_job=None)),
        ('manage', dict(opt='listprojects', project=None, version_spider_job=None))
    ])

    # files
    from .files.log import LogView
    register_view(LogView, 'log', [('log/<opt>/<project>/<spider>/<job>', None)])

    from .files.logs import LogsView
    register_view(LogsView, 'logs', [
        ('logs/<project>/<spider>', None),
        ('logs/<project>', dict(spider=None)),
        ('logs', dict(project=None, spider=None))
    ])

    from .files.items import ItemsView
    register_view(ItemsView, 'items', [
        ('items/<project>/<spider>', None),
        ('items/<project>', dict(spider=None)),
        ('items', dict(project=None, spider=None))
    ])

    from .files.parse import UploadLogView, UploadedLogView
    register_view(UploadLogView, 'parse.upload', [('parse/upload', None)])
    register_view(UploadedLogView, 'parse.uploaded', [('parse/uploaded/<filename>', None)])

    from .files.parse import bp as bp_parse_source
    app.register_blueprint(bp_parse_source)

    # system
    from .system.settings import SettingsView
    register_view(SettingsView, 'settings', [('settings', None)])



# coding: utf8
import re

from flask import render_template, url_for

from ..myview import MyView
from ..vars import DEFAULT_LATEST_VERSION, pattern_directory, keys_directory


class LogsView(MyView):
    methods = ['GET']

    def __init__(self):
        super(self.__class__, self).__init__()

        self.project = self.view_args['project']
        self.spider = self.view_args['spider']

        self.url = 'http://{}/logs/{}{}'.format(self.SCRAPYD_SERVER,
                                                '%s/' % self.project if self.project else '',
                                                '%s/' % self.spider if self.spider else '')
        self.template = 'scrapydweb/logs.html'
        self.text = ''

    def dispatch_request(self, **kwargs):
        status_code, self.text = self.make_request(self.url, api=False, auth=self.AUTH)
        if status_code != 200 or not re.search(r'Directory listing for /logs/', self.text):
            kwargs = dict(
                node=self.node,
                url=self.url,
                status_code=status_code,
                text=self.text
            )
            return render_template(self.template_fail, **kwargs)

        return self.generate_response()

    def generate_response(self):
        rows = [dict(zip(keys_directory, row)) for row in pattern_directory.findall(self.text)]

        for row in rows:
            if self.project and self.spider:
                # <a href="098726cca42b11e8a8b514dda9e91c2f.log">098726cca42b11e8a8b514dda9e91c2f.log</a>
                m = re.search(r'>(.*?)<', row['filename'])
                filename = m.group(1)
                row['filename'] = u'<a class="link" target="_blank" href="{}">{}</a>'.format(
                                   self.url + filename, filename)

                row['url_log_utf8'] = url_for('log', node=self.node, opt='utf8',
                                              project=self.project, spider=self.spider,
                                              job=filename, with_ext='True')
                row['url_log_stats'] = url_for('log', node=self.node, opt='stats',
                                               project=self.project, spider=self.spider,
                                               job=filename, with_ext='True')

                row['url_start'] = url_for('schedule.schedule', node=self.node, project=self.project,
                                           version=DEFAULT_LATEST_VERSION, spider=self.spider)
                row['url_multinode_start'] = url_for('overview', node=self.node,
                                                     opt='schedule', project=self.project,
                                                     version_job=DEFAULT_LATEST_VERSION, spider=self.spider)
            else:
                # <a href="proxy/">proxy/</a>
                row['filename'] = re.sub(r'href=', 'class="link" href=', row['filename'])

        kwargs = dict(
            node=self.node,
            project=self.project,
            spider=self.spider,
            url=self.url,
            rows=rows
        )
        return render_template(self.template, **kwargs)



# coding: utf8
import re
import time

from .myview import MyView
from .vars import DEFAULT_LATEST_VERSION


API_MAP = dict(start='schedule', stop='cancel', forcestop='cancel')


class ApiView(MyView):

    def __init__(self):
        super(self.__class__, self).__init__()

        self.opt = self.view_args['opt']
        self.project = self.view_args['project']
        self.version_spider_job = self.view_args['version_spider_job']

        self.url = 'http://{}/{}.json'.format(self.SCRAPYD_SERVER, API_MAP.get(self.opt, self.opt))
        self.data = None

    def dispatch_request(self, **kwargs):
        # if self.node < 1 or self.node > len(self.SCRAPYD_SERVERS):
            # message = 'node index %s error, which should be between 1 and %s' % (self.node, len(self.SCRAPYD_SERVERS))
            # self.logger.error('!!!!! %s' % message)
            # return self.json_dumps(dict(status_code=-1, status='error', message=message), sort_keys=False)

        self.update_url()
        self.update_data()
        return self.generate_response()

    def update_url(self):
        if self.opt in ['listversions', 'listjobs']:
            self.url += '?project=%s' % self.project
        elif self.opt == 'listspiders':
            if self.version_spider_job == DEFAULT_LATEST_VERSION:
                self.url += '?project=%s' % self.project
            else:
                # Should be _version
                self.url += '?project=%s&_version=%s' % (self.project, self.version_spider_job)

    def update_data(self):
        self.data = dict(project=self.project)
        if self.opt == 'start':
            self.data['spider'] = self.version_spider_job
            self.data['jobid'] = self.get_now_string()
        elif self.opt in ['stop', 'forcestop']:
            self.data['job'] = self.version_spider_job
        elif self.opt == 'delversion':
            self.data['version'] = self.version_spider_job
        elif self.opt == 'delproject':
            pass
        else:
            self.data = None

    def generate_response(self):
        timeout = 5 if self.opt == 'daemonstatus' else 60
        times = 2 if self.opt == 'forcestop' else 1
        js = {}
        for __ in range(times):
            status_code, js = self.make_request(self.url, self.data, timeout=timeout, auth=self.AUTH)
            if times != 1:
                js['times'] = times
                time.sleep(2)

        if js['status'] != 'ok':
            if self.opt == 'listversions':
                js['info'] = (
                    "Maybe it's caused by failing to compare versions, "
                    "you can check out the INFO section in the Deploy page, "
                    "and solve the problem in the Manage page."
                )
            elif self.opt == 'listspiders' or re.search('no active project', js.get('message', '')):
                js['info'] = "Maybe the project has been deleted, check out the Manage page"

        return self.json_dumps(js)



# coding: utf8
from flask import render_template, flash, url_for

from ..myview import MyView
from ..vars import CHECK_UPDATE, INFO, WARN


page_view = 0 if CHECK_UPDATE else 1


class OverviewView(MyView):

    def __init__(self):
        super(self.__class__, self).__init__()

        self.opt = self.view_args['opt']
        self.project = self.view_args['project']
        self.version_job = self.view_args['version_job']
        self.spider = self.view_args['spider']

        self.url = 'http://%s/daemonstatus.json' % self.SCRAPYD_SERVER
        self.template = 'scrapydweb/overview.html'
        self.selected_nodes = []

        self.flag = ''
        if len(self.SCRAPYD_SERVERS) > 1:
            self.flag += 'A' if self.ENABLE_AUTH else '-'
            self.flag += 'C' if self.ENABLE_CACHE else '-'
            self.flag += 'E' if self.ENABLE_EMAIL else '-'

    def dispatch_request(self, **kwargs):
        global page_view
        page_view += 1

        if len(self.SCRAPYD_SERVERS) == 1:
            # flash("Run ScrapydWeb with argument '-ss 127.0.0.1 -ss username:password@192.168.123.123:6801#group' "
                  # "to set up more than one Scrapyd server to control.", INFO)
            pass
        else:
            if not self.ENABLE_AUTH:
                flash("Set 'ENABLE_AUTH = True' to enable basic auth for web UI", INFO)
            if not self.ENABLE_CACHE:
                flash("Set 'ENABLE_CACHE = True' to enable caching HTML for Log and Stats page", WARN)
            if not self.ENABLE_EMAIL:
                flash("Set 'ENABLE_EMAIL = True' to enable email notice", WARN)

        if self.POST:
            self.selected_nodes = self.get_selected_nodes()
        else:
            if len(self.SCRAPYD_SERVERS) == 1:
                self.selected_nodes = [1]
            else:
                self.selected_nodes = []

        kwargs = dict(
            node=self.node,
            opt=self.opt,
            project=self.project,
            version_job=self.version_job,
            spider=self.spider,
            url=self.url,
            selected_nodes=self.selected_nodes,
            IS_IE_EDGE=self.IS_IE_EDGE,
            flag=self.flag,
            page_view=page_view,
            url_daemonstatus=url_for('api', node=self.node, opt='daemonstatus'),
            url_listprojects=url_for('api', node=self.node, opt='listprojects'),
            url_listversions=url_for('api', node=self.node, opt='listversions', project='PROJECT_PLACEHOLDER'),
            url_listspiders=url_for('api', node=self.node, opt='listspiders', project='PROJECT_PLACEHOLDER',
                                    version_spider_job='VERSION_PLACEHOLDER'),
            url_listjobs=url_for('api', node=self.node, opt='listjobs', project='PROJECT_PLACEHOLDER'),
            url_deploy=url_for('deploy.deploy', node=self.node),
            url_schedule=url_for('schedule.schedule', node=self.node, project='PROJECT_PLACEHOLDER',
                                 version='VERSION_PLACEHOLDER', spider='SPIDER_PLACEHOLDER'),
            url_stop=url_for('multinode', node=self.node, opt='stop', project='PROJECT_PLACEHOLDER',
                             version_job='JOB_PLACEHOLDER'),
            url_delversion=url_for('multinode', node=self.node, opt='delversion', project='PROJECT_PLACEHOLDER',
                                   version_job='VERSION_PLACEHOLDER'),
            url_delproject=url_for('multinode', node=self.node, opt='delproject', project='PROJECT_PLACEHOLDER')
        )
        return render_template(self.template, **kwargs)



# coding: utf8
import os
import io
import time
import glob
import re


CWD = os.path.dirname(os.path.abspath(__file__))
DATA_PATH = os.path.join(CWD, 'data')
PARSE_PATH = os.path.join(DATA_PATH, 'parse')
CACHE_PATH = os.path.join(DATA_PATH, 'cache')
DEPLOY_PATH = os.path.join(DATA_PATH, 'deploy')
SCHEDULE_PATH = os.path.join(DATA_PATH, 'schedule')
DEMO_PROJECTS_PATH = os.path.join(DATA_PATH, 'demo_projects')

for p in [DATA_PATH, PARSE_PATH, CACHE_PATH, DEPLOY_PATH, SCHEDULE_PATH, DEMO_PROJECTS_PATH]:
    if not os.path.isdir(p):
        os.mkdir(p)
    elif p in [PARSE_PATH, DEPLOY_PATH, SCHEDULE_PATH]:
        for file in glob.glob(os.path.join(p, '*.*')):
            if not os.path.split(file)[-1] in ['demo.txt', 'history.log']:
                os.remove(file)


LAST_CHECK_UPDATE = os.path.join(DATA_PATH, 'last_check_update')

try:
    if not os.path.exists(LAST_CHECK_UPDATE):
        with io.open(LAST_CHECK_UPDATE, 'w') as f:
            f.write(u'{:.0f}'.format(time.time()))
        CHECK_UPDATE = True
    else:
        with io.open(LAST_CHECK_UPDATE, 'r+') as f:
            if time.time() - int(f.read()) > 3600 * 24 * 7:
                f.seek(0)
                f.write(u'{:.0f}'.format(time.time()))
                CHECK_UPDATE = True
            else:
                CHECK_UPDATE = False
except:
    try:
        os.remove(LAST_CHECK_UPDATE)
    except:
        pass
    CHECK_UPDATE = True


INFO = 'info'
WARN = 'warning'
DEFAULT_LATEST_VERSION = "default: the latest version"


UA_DICT = {
    'chrome': ("Mozilla/5.0 (Windows NT 6.1; Win64; x64) "
               "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/63.0.3239.84 Safari/537.36"),
    'iOS': ("Mozilla/5.0 (iPhone; CPU iPhone OS 10_3 like Mac OS X) "
            "AppleWebKit/602.1.50 (KHTML, like Gecko) CriOS/56.0.2924.75 Mobile/14E5239e Safari/602.1"),
    'iPad': ("Mozilla/5.0 (iPad; CPU OS 9_1 like Mac OS X) "
             "AppleWebKit/601.1.46 (KHTML, like Gecko) Version/9.0 Mobile/13B143 Safari/601.1"),
    'Android': ("Mozilla/5.0 (Linux; Android 5.1.1; Nexus 6 Build/LYZ28E) "
                "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/63.0.3239.84 Mobile Safari/537.36")
}


# For log.py
ALLOWED_SCRAPYD_LOG_EXTENSIONS = ['.log', '.log.gz', '.txt', '.gz', '']


# For run.py
# r'^(?:(?:(.*?)\:)(?:(.*?)@))?(.*?)(?:\:(.*?))?(?:#(.*?))?$'
pattern_scrapyd_server = re.compile(r"""
                        ^
                        (?:
                            (?:(.*?):)     # username:
                            (?:(.*?)@)      # password@
                        )?
                        (.*?)               # ip
                        (?::(.*?))?        # :port
                        (?:\#(.*?))?        # #group
                        $
                    """, re.X)


# For dashboard
pattern_jobs = re.compile(r"""<tr>
                        <td>(?P<Project>.*?)</td>
                        <td>(?P<Spider>.*?)</td>
                        <td>(?P<Job>.*?)</td>
                        (?:<td>(?P<PID>.*?)</td>)?
                        (?:<td>(?P<Start>.*?)</td>)?
                        (?:<td>(?P<Runtime>.*?)</td>)?
                        (?:<td>(?P<Finish>.*?)</td>)?
                        (?:<td>(?P<Log>.*?)</td>)?
                        (?:<td>(?P<Items>.*?)</td>)?
                        </tr>
                    """, re.X)
keys_jobs = ['project', 'spider', 'job', 'pid', 'start', 'runtime', 'finish', 'log', 'items']


# For directory
pattern_directory = re.compile(r"""<tr\sclass="(?P<odd_even>odd|even)">\n
                                   \s+<td>(?P<filename>.*?)</td>\n
                                   \s+<td>(?P<size>.*?)</td>\n
                                   \s+<td>(?P<content_type>.*?)</td>\n
                                   \s+<td>(?P<content_encoding>.*?)</td>\n
                                </tr>
                                """, re.X)

keys_directory = ['odd_even', 'filename', 'size', 'content_type', 'content_encoding']


# For email notice
EMAIL_CONTENT_KEYS = [
    'log_critical_count',
    'log_error_count',
    'log_warning_count',
    'log_redirect_count',
    'log_retry_count',
    'log_ignore_count',
    'crawled_pages',
    'scraped_items'
]

EMAIL_TRIGGER_KEYS = ['CRITICAL', 'ERROR', 'WARNING', 'REDIRECT', 'RETRY', 'IGNORE']



# coding: utf8
import os
import sys
import io
import time
from datetime import datetime, date
import re
from collections import OrderedDict
from subprocess import Popen

from flask import render_template, url_for, request, send_from_directory, flash

from ..myview import MyView
from ..vars import CACHE_PATH, INFO, WARN, EMAIL_CONTENT_KEYS, EMAIL_TRIGGER_KEYS
from .utils import parse_log


CWD = os.path.dirname(os.path.abspath(__file__))
job_data_dict = {}
# job_finished_set would be updated by caching POST with ?job_finished=True > email_notice() only,
# used for determining whether to show 'refresh' button with '?refresh_cache=True'.
job_finished_set = set()


# http://flask.pocoo.org/docs/1.0/api/#flask.views.View
# http://flask.pocoo.org/docs/1.0/views/
class LogView(MyView):

    def __init__(self):
        super(self.__class__, self).__init__()  # super().__init__()

        self.opt = self.view_args['opt']
        self.project = self.view_args['project']
        self.spider = self.view_args['spider']
        self.job = self.view_args['job']

        self.job_data_dict = job_data_dict
        self.job_finished_set = job_finished_set

        self.url = u'http://{}/logs/{}/{}/{}.log'.format(self.SCRAPYD_SERVER, self.project, self.spider, self.job)
        self.postfix = '_mobileui' if self.IS_MOBILEUI else ''
        self.template_stats = 'scrapydweb/stats.html'
        self.template_stats_mobileui = 'scrapydweb/stats_mobileui.html'
        self.template_utf8 = 'scrapydweb/utf8.html'
        self.template_utf8_mobileui = 'scrapydweb/utf8_mobileui.html'
        self.status_code = 0
        self.text = ''
        self.spider_path = ''
        self.stats_kwargs = {}
        self.refresh_url = ''
        self.refresh_url_mobileui = ''
        self.stats_html = ''
        self.stats_mobileui_html = ''
        self.utf8_html = ''
        self.utf8_mobileui_html = ''

        # ([0] * 8, [False] * 6, False, time.time())
        self.job_stats_previous = []
        self.triggered_list = []
        self.has_been_stopped = False
        self.last_send_timestamp = 0
        self.job_stats = []
        self.job_stats_diff = []
        self.email_content_kwargs = {}
        self.flag = ''

        # In Logs page: xxx.log/?with_ext=True
        self.with_ext = request.args.get('with_ext', None)
        if self.with_ext:
            self.SCRAPYD_LOG_EXTENSIONS = ['']

        self.job_key = '/%s/%s/%s/%s' % (self.node, self.project, self.spider, self.job)

        # request from caching would post ?job_finished=True for finished job,
        # as well as links of finished job in Dashboard page
        self.job_finished = request.args.get('job_finished', None)

        # Three conditions in total
        # self.disable_cache = self.IS_MOBILEUI or not self.ENABLE_CACHE
        _condition = request.args.get('refresh_cache') or self.POST  # scrapydweb/utils/cache.py would use 'POST'
        self.to_refresh_cache = True if self.ENABLE_CACHE and _condition else False
        self.to_read_cache = True if self.ENABLE_CACHE and not self.to_refresh_cache else False

    def dispatch_request(self, **kwargs):
        if self.ENABLE_CACHE:
            self.mkdir_spider_path()

        # Return cached html
        if self.to_read_cache:
            htmlname = '%s_%s%s.html' % (self.job, self.opt, self.postfix)
            if os.path.exists(os.path.join(self.spider_path, htmlname)):
                return send_from_directory(self.spider_path, htmlname)

        # Try to read local scrapy log file
        if self.SCRAPYD_SERVER.split(':')[0] == '127.0.0.1' and self.SCRAPYD_LOGS_DIR:
            self.read_local_scrapy_log()

        # Has to request scrapy log
        if not self.text:
            self.request_scrapy_log()
            if self.status_code != 200:
                return render_template(self.template_fail, node=self.node,
                                       url=self.url, status_code=self.status_code, text=self.text)

        self.render_stats_html()
        self.render_utf8_html()
        if self.ENABLE_CACHE:
            self.save_html()

        if self.ENABLE_EMAIL and self.POST:  # only cache.py would make POST request
            self.email_notice()

        return getattr(self, '%s%s_html' % (self.opt, self.postfix))

    def mkdir_spider_path(self):
        node_path = os.path.join(CACHE_PATH, re.sub(r'[.:]', '_', self.SCRAPYD_SERVER))
        project_path = os.path.join(node_path, self.project)
        self.spider_path = os.path.join(project_path, self.spider)

        if not os.path.isdir(CACHE_PATH):
            os.mkdir(CACHE_PATH)
        if not os.path.isdir(node_path):
            os.mkdir(node_path)
        if not os.path.isdir(project_path):
            os.mkdir(project_path)
        if not os.path.isdir(self.spider_path):
            os.mkdir(self.spider_path)

    def read_local_scrapy_log(self):
        file_found = False
        logfile = ''
        for ext in self.SCRAPYD_LOG_EXTENSIONS:
            logfile = os.path.join(self.SCRAPYD_LOGS_DIR, self.project, self.spider, self.job + ext)
            if os.path.exists(logfile):
                with io.open(logfile, 'r', encoding='utf-8', errors='ignore') as f:
                    self.text = f.read()
                flash("Using local scrapy logfile: %s" % logfile, INFO)
                file_found = True
                break
        if not file_found:
            flash("Local logfile '%s' NOT found, making request to Scrapyd server instead" % logfile, WARN)

    def request_scrapy_log(self):
        url_without_ext = re.sub(r'.log$', '', self.url)

        for ext in self.SCRAPYD_LOG_EXTENSIONS:
            self.url = '%s%s' % (url_without_ext, ext)
            self.status_code, self.text = self.make_request(self.url, api=False, auth=self.AUTH)
            if self.status_code == 200:
                break
        else:
            self.url = '%s%s' % (url_without_ext, self.SCRAPYD_LOG_EXTENSIONS[0])

    def render_stats_html(self):
        self.stats_kwargs = dict(
            project=self.project,
            spider=self.spider,
            job=self.job,
            url_source=self.url,
            url_utf8=url_for('log', node=self.node, opt='utf8', project=self.project, spider=self.spider,
                             job=self.job, with_ext=self.with_ext, job_finished=self.job_finished),
            url_utf8_mobile=url_for('log', node=self.node, opt='utf8', project=self.project,
                                    spider=self.spider, job=self.job, ui='mobile'),
            url_dashboard_mobile=url_for('dashboard', node=self.node, ui='mobile'),
            LAST_LOG_ALERT_SECONDS=self.LAST_LOG_ALERT_SECONDS
        )

        parse_log(self.text, self.stats_kwargs)

        # To show 'refresh' button with '?refresh_cache=True'
        if (self.ENABLE_CACHE
           and not self.stats_kwargs.get('finish_reason')
           and not self.job_finished
           and self.job_key not in self.job_finished_set):
            # http://flask.pocoo.org/docs/1.0/api/#flask.Request.url_root
            # _query_string = '?refresh_cache=True'
            # self.refresh_url = request.script_root + request.path + _query_string
            # For Log and stats buttons in the Logs page: with_ext=self.with_ext
            self.refresh_url = url_for('log', node=self.node, opt='stats', project=self.project, spider=self.spider,
                                       job=self.job, refresh_cache='True', with_ext=self.with_ext)
            self.refresh_url_mobileui = self.refresh_url + '&ui=mobile'

        self.stats_kwargs['refresh_url'] = self.refresh_url
        self.stats_kwargs['refresh_url_mobileui'] = self.refresh_url_mobileui
        self.stats_kwargs['last_refresh_timestamp'] = time.time()
        self.stats_html = render_template(self.template_stats, node=self.node, **self.stats_kwargs)

        self.stats_kwargs['IS_MOBILE'] = True
        self.stats_mobileui_html = render_template(self.template_stats_mobileui, node=self.node, **self.stats_kwargs)

    def render_utf8_html(self):
        utf8_kwargs = dict(
            node=self.node,
            project=self.project,
            spider=self.spider,
            url_source=self.url,
            url_stats=url_for('log', node=self.node, opt='stats', project=self.project, spider=self.spider,
                              job=self.job, with_ext=self.with_ext, job_finished=self.job_finished),
            url_stats_mobile=url_for('log', node=self.node, opt='stats', project=self.project,
                                     spider=self.spider, job=self.job, ui='mobile'),
            url_dashboard_mobile=url_for('dashboard', node=self.node, ui='mobile'),
            refresh_url=re.sub(r'(/\d+/log/)stats/', r'\1utf8/', self.refresh_url, count=1),
            refresh_url_mobileui=re.sub(r'(/\d+/log/)stats/', r'\1utf8/', self.refresh_url_mobileui, count=1),
            last_refresh_timestamp=time.time(),
            LAST_LOG_ALERT_SECONDS=self.LAST_LOG_ALERT_SECONDS,
            text=self.text
        )

        self.utf8_html = render_template(self.template_utf8, **utf8_kwargs)

        utf8_kwargs['IS_MOBILE'] = True
        self.utf8_mobileui_html = render_template(self.template_utf8_mobileui, **utf8_kwargs)

    def save_html(self):
        for s in ['stats', 'stats_mobileui', 'utf8', 'utf8_mobileui']:
            html = getattr(self, '%s_html' % s)
            htmlname = '%s_%s.html' % (self.job, s)
            htmlpath = os.path.join(self.spider_path, htmlname)
            try:
                with io.open(htmlpath, 'w', encoding='utf-8', errors='ignore') as f:
                    f.write(html)
            except Exception as err:
                self.logger.error("Fail to cache HTML to %s: %s" % (htmlpath, err))
                try:
                    os.remove(htmlpath)
                except:
                    pass

    def email_notice(self):
        job_data_default = ([0] * 8, [False] * 6, False, time.time())
        job_data = self.job_data_dict.setdefault(self.job_key, job_data_default)
        (self.job_stats_previous, self.triggered_list, self.has_been_stopped, self.last_send_timestamp) = job_data
        # print(self.job_data_dict)
        self.job_stats = [self.stats_kwargs['re_matches'][i]['count'] for i in range(6)]
        self.job_stats.extend([self.stats_kwargs['crawled_pages'], self.stats_kwargs['scraped_items']])
        self.job_stats_diff = [j - i for i, j in zip(self.job_stats_previous, self.job_stats)]

        self.set_email_content_kwargs()
        self.set_email_flag()
        self.handle_email_flag()

    def set_email_content_kwargs(self):
        # For compatibility with Python 2, use OrderedDict() to keep insertion order
        self.email_content_kwargs = OrderedDict()
        self.email_content_kwargs['SCRAPYD_SERVER'] = self.SCRAPYD_SERVER
        self.email_content_kwargs['project'] = self.stats_kwargs['project']
        self.email_content_kwargs['spider'] = self.stats_kwargs['spider']
        self.email_content_kwargs['job'] = self.stats_kwargs['job']
        self.email_content_kwargs['start_time'] = self.stats_kwargs['first_log_time']
        self.email_content_kwargs['latest_log_time'] = self.stats_kwargs['latest_log_time']
        self.email_content_kwargs['elasped'] = self.stats_kwargs['elasped']
        self.email_content_kwargs['finish_reason'] = self.stats_kwargs['finish_reason'] or 'NOT found in log'
        self.email_content_kwargs['url_stats'] = request.url + '&ui=mobile'  # cache.py would POST '?job_finished='

        for idx, i in enumerate(EMAIL_CONTENT_KEYS):
            if self.job_stats_diff[idx]:
                self.email_content_kwargs[i] = '%s + %s' % (self.job_stats_previous[idx], self.job_stats_diff[idx])
            else:
                self.email_content_kwargs[i] = self.job_stats[idx]

        _bind = '127.0.0.1' if self.SCRAPYDWEB_BIND == '0.0.0.0' else self.SCRAPYDWEB_BIND
        _url_stop = url_for('api', node=self.node, opt='stop', project=self.project, version_spider_job=self.job)
        self.email_content_kwargs['url_stop'] = 'http://%s:%s%s' % (_bind, self.SCRAPYDWEB_PORT, _url_stop)

        now_timestamp = time.time()
        latest_crawl_diff = now_timestamp - self.stats_kwargs['latest_crawl_timestamp']
        latest_scrape_diff = now_timestamp - self.stats_kwargs['latest_scrape_timestamp']
        latest_log_diff = now_timestamp - self.stats_kwargs['latest_log_timestamp']
        self.email_content_kwargs['latest_crawl'] = "%s seconds ago" % int(latest_crawl_diff)
        self.email_content_kwargs['latest_scrape'] = "%s seconds ago" % int(latest_scrape_diff)
        self.email_content_kwargs['latest_log'] = "%s seconds ago" % int(latest_log_diff)
        self.email_content_kwargs['current_time'] = time.ctime()

    def set_email_flag(self):
        if self.ON_JOB_FINISHED and self.job_finished:
            self.flag = 'Finished'
        elif not all(self.triggered_list):
            to_forcestop = False
            to_stop = False
            # The order of the elements in EMAIL_TRIGGER_KEYS matters:
            # ['CRITICAL', 'ERROR', 'WARNING', 'REDIRECT', 'RETRY', 'IGNORE']
            for idx, key in enumerate(EMAIL_TRIGGER_KEYS):
                if (0 < getattr(self, 'LOG_%s_THRESHOLD' % key, 0) <= self.job_stats[idx]
                   and not self.triggered_list[idx]):
                    self.triggered_list[idx] = True
                    self.email_content_kwargs['log_%s_count' % key.lower()] += ' triggered!!!'
                    if getattr(self, 'LOG_%s_TRIGGER_FORCESTOP' % key):
                        self.flag = '%s_ForceStop' % key if '_ForceStop' not in self.flag else self.flag
                        to_forcestop = True
                    elif getattr(self, 'LOG_%s_TRIGGER_STOP' % key) and not self.has_been_stopped:
                        self.flag = '%s_Stop' % key if 'Stop' not in self.flag else self.flag
                        self.has_been_stopped = True  # Execute 'Stop' one time at most to avoid unclean shutdown
                        to_stop = True
                    elif not self.has_been_stopped:
                        self.flag = '%s_Trigger' % key if not self.flag else self.flag
            if to_forcestop:
                self.logger.debug("%s: %s" % (self.flag, self.job_key))
                # api(self.node, 'forcestop', self.project, self.job)
                _url = url_for('api', node=self.node, opt='forcestop',
                               project=self.project, version_spider_job=self.job)
                self.get_response_from_view(_url)
            elif to_stop:
                self.logger.debug("%s: %s" % (self.flag, self.job_key))
                # api(self.node, 'stop', self.project, self.job)
                _url = url_for('api', node=self.node, opt='stop',
                               project=self.project, version_spider_job=self.job)
                self.get_response_from_view(_url)

        if not self.flag and 0 < self.ON_JOB_RUNNING_INTERVAL <= time.time() - self.last_send_timestamp:
            self.flag = 'Running'

    def handle_email_flag(self):
        if self.flag:
            # Send email
            now_day = date.isoweekday(datetime.now())
            now_hour = datetime.now().hour
            if now_day in self.EMAIL_WORKING_DAYS and now_hour in self.EMAIL_WORKING_HOURS:
                self.EMAIL_KWARGS['subject'] = '%s %s #scrapydweb' % (self.flag, self.job_key)
                self.EMAIL_KWARGS['content'] = self.json_dumps(self.email_content_kwargs, sort_keys=False)

                args = [
                    sys.executable,
                    os.path.join(os.path.dirname(CWD), 'utils', 'send_email.py'),
                    self.json_dumps(self.EMAIL_KWARGS)
                ]
                self.logger.info("Sending email: %s" % self.EMAIL_KWARGS['subject'])
                Popen(args)

            # Update self.job_data_dict (last_send_timestamp would be updated only when flag is non-empty)
            self.logger.debug(self.job_data_dict[self.job_key])
            self.job_data_dict[self.job_key] = (self.job_stats, self.triggered_list, self.has_been_stopped, time.time())
            self.logger.debug(self.job_data_dict[self.job_key])

        if self.job_finished:
            self.job_data_dict.pop(self.job_key)
            if len(self.job_finished_set) > 1000:
                self.job_finished_set.clear()
            self.job_finished_set.add(self.job_key)



# coding: utf8
"""
source: https://github.com/scrapy/scrapyd-client
"""
import os
import sys
import glob
import tempfile
from shutil import copyfile
from subprocess import check_call
import errno
try:
    from ConfigParser import SafeConfigParser
except ImportError:
    from configparser import ConfigParser as SafeConfigParser

from flask import current_app as app


_SETUP_PY_TEMPLATE = """# Automatically created by: scrapydweb x scrapyd-client

from setuptools import setup, find_packages

setup(
    name         = 'project',
    version      = '1.0',
    packages     = find_packages(),
    entry_points = {'scrapy': ['settings = %(settings)s']},
)
"""


def get_config(sources):
    """Get Scrapy config file as a SafeConfigParser"""
    # sources = get_sources(use_closest)
    cfg = SafeConfigParser()
    cfg.read(sources)
    return cfg


def retry_on_eintr(func, *args, **kw):
    """Run a function and retry it while getting EINTR errors"""
    while True:
        try:
            return func(*args, **kw)
        except IOError as e:
            if e.errno != errno.EINTR:
                raise


def _build_egg(scrapy_cfg_path):
    cwd = os.getcwd()
    os.chdir(os.path.dirname(scrapy_cfg_path))

    if os.path.exists('setup.py'):
        copyfile('setup.py', 'setup_backup.py')
    # Content in myproject/scrapy.cfg: [settings] default = demo.settings
    settings = get_config(scrapy_cfg_path).get('settings', 'default')  # demo.settings
    _create_default_setup_py(settings=settings)

    d = tempfile.mkdtemp(prefix="scrapydweb-deploy-")
    o = open(os.path.join(d, "stdout"), "wb")
    e = open(os.path.join(d, "stderr"), "wb")
    try:
        retry_on_eintr(check_call, [sys.executable, 'setup.py', 'clean', '-a', 'bdist_egg', '-d', d],
                       stdout=o, stderr=e)
    except:
        raise
    else:
        egg = glob.glob(os.path.join(d, '*.egg'))[0]
    finally:
        os.chdir(cwd)
        o.close()
        e.close()

    return egg, d


def _create_default_setup_py(**kwargs):
    with open('setup.py', 'w') as f:
        content = _SETUP_PY_TEMPLATE % kwargs
        app.logger.debug('New setup.py')
        # app.logger.debug(content)
        f.write(content)



# coding: utf8
import re
try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin

from flask import render_template, flash, url_for

from ..myview import MyView
from ..vars import CHECK_UPDATE, INFO, WARN, pattern_jobs, keys_jobs, DEFAULT_LATEST_VERSION


page_view = 0 if CHECK_UPDATE else 1


class DashboardView(MyView):
    methods = ['GET']

    def __init__(self):
        super(self.__class__, self).__init__()

        self.url = 'http://%s/jobs' % self.SCRAPYD_SERVER
        self.template = 'scrapydweb/dashboard_mobileui.html' if self.IS_MOBILEUI else 'scrapydweb/dashboard.html'
        self.text = ''

        self.flag = ''
        if not self.IS_MOBILEUI and len(self.SCRAPYD_SERVERS) == 1:
            self.flag += 'A' if self.ENABLE_AUTH else '-'
            self.flag += 'C' if self.ENABLE_CACHE else '-'
            self.flag += 'E' if self.ENABLE_EMAIL else '-'

    def dispatch_request(self, **kwargs):
        global page_view
        page_view += 1

        status_code, self.text = self.make_request(self.url, api=False, auth=self.AUTH)

        if status_code != 200 or not re.search(r'Jobs', self.text):
            message = 'Click the above link to make sure your Scrapyd server is accessable'
            return render_template(self.template_fail, node=self.node,
                                   url=self.url, status_code=status_code,
                                   text=self.text, message=message)

        return self.generate_response()

    def generate_response(self):
        if self.IS_MOBILEUI:
            # flash("<a href='/'>Visit desktop version</a> to experience full features.", INFO)
            pass
        else:
            if len(self.SCRAPYD_SERVERS) > 1 and page_view == 1:
                flash("Use the navigation buttons above to quick scan the same page of neighbouring node", INFO)
            if not self.ENABLE_AUTH and len(self.SCRAPYD_SERVERS) == 1:
                flash("Set 'ENABLE_AUTH = True' to enable basic auth for web UI", INFO)
            if not self.SCRAPYD_LOGS_DIR and self.SCRAPYD_SERVER.split(':')[0] == '127.0.0.1':
                flash("Set up the 'SCRAPYD_LOGS_DIR' item to speed up loading scrapy logs.", INFO)
            if not self.ENABLE_CACHE:
                flash("Set 'ENABLE_CACHE = True' to enable caching HTML for Log and Stats page", WARN)
            if not self.ENABLE_EMAIL and len(self.SCRAPYD_SERVERS) == 1:
                flash("Set 'ENABLE_EMAIL = True' to enable email notice", WARN)

        rows = [dict(zip(keys_jobs, row)) for row in pattern_jobs.findall(self.text)]
        pending_rows = []
        running_rows = []
        finished_rows = []
        for row in rows:
            if not row['start']:
                pending_rows.append(row)
            else:
                job_finished = 'True' if row['finish'] else None
                row['url_utf8'] = url_for('log', node=self.node, opt='utf8', project=row['project'], ui=self.UI,
                                          spider=row['spider'], job=row['job'], job_finished=job_finished)
                row['url_stats'] = url_for('log', node=self.node, opt='stats', project=row['project'], ui=self.UI,
                                           spider=row['spider'], job=row['job'], job_finished=job_finished)
                row['url_multinode_stop'] = url_for('overview', node=self.node, opt='stop', project=row['project'],
                                                    version_job=row['job'])
                row['url_stop'] = url_for('api', node=self.node, opt='stop', project=row['project'],
                                          version_spider_job=row['job'])
                row['url_forcestop'] = url_for('api', node=self.node, opt='forcestop', project=row['project'],
                                               version_spider_job=row['job'])
                row['url_multinode_run'] = url_for('overview', node=self.node, opt='schedule', project=row['project'],
                                                   version_job=DEFAULT_LATEST_VERSION, spider=row['spider'])
                row['url_schedule'] = url_for('schedule.schedule', node=self.node, project=row['project'],
                                              version=DEFAULT_LATEST_VERSION, spider=row['spider'])
                row['url_start'] = url_for('api', node=self.node, opt='start', project=row['project'],
                                           version_spider_job=row['spider'])
                # <a href='/items/demo/test/2018-10-12_205507.jl'>Items</a>
                if row['items']:
                    _url_items = re.search(r"href='(.*?)'>", row['items']).group(1)
                    # row['url_items'] = re.sub(r'/jobs$', _url_items, self.url) + _url_items
                    row['url_items'] = urljoin(self.url, _url_items)

                if row['finish']:
                    finished_rows.append(row)
                else:
                    running_rows.append(row)

        kwargs = dict(
            node=self.node,
            colspan=12,
            url=self.url,
            scrapydweb_url='http://%s:%s' % (self.SCRAPYDWEB_BIND, self.SCRAPYDWEB_PORT),
            pending_rows=pending_rows,
            running_rows=running_rows,
            finished_rows=finished_rows,
            SHOW_DASHBOARD_JOB_COLUMN=self.SHOW_DASHBOARD_JOB_COLUMN,
            DASHBOARD_RELOAD_INTERVAL=self.DASHBOARD_RELOAD_INTERVAL,
            IS_IE_EDGE=self.IS_IE_EDGE,
            flag=self.flag,
            page_view=page_view
        )
        return render_template(self.template, **kwargs)



# coding: utf8
import os
import sys
import argparse
from shutil import copyfile
from multiprocessing.dummy import Pool as ThreadPool

import requests
from flask import request

# from . import create_app  # --debug: ImportError: cannot import name 'create_app'
from scrapydweb import create_app
from scrapydweb.__version__ import __version__, __description__
from scrapydweb.vars import pattern_scrapyd_server, LAST_CHECK_UPDATE
from scrapydweb.utils.utils import printf, find_scrapydweb_settings_py, authenticate
from scrapydweb.utils.check_app_config import check_app_config
from scrapydweb.utils.init_caching import init_caching


CWD = os.path.dirname(os.path.abspath(__file__))
SCRAPYDWEB_SETTINGS_PY = 'scrapydweb_settings_v5.py'
scrapydweb_settings_py_path = os.path.join(os.getcwd(), SCRAPYDWEB_SETTINGS_PY)


def main():
    main_pid = os.getpid()
    printf("Main pid: %s" % main_pid)
    printf("ScrapydWeb version: %s" % __version__)
    printf("Use the 'scrapydweb -h' command to get help")
    printf("Loading default settings from %s" % os.path.join(CWD, 'default_settings.py'))

    app = create_app()
    load_custom_config(app.config)

    args = parse_args(app.config)
    # "scrapydweb -h" ends up here
    update_app_config(app.config, args)
    # from pprint import pprint
    # pprint(app.config)
    try:
        check_app_config(app.config)
    except AssertionError as err:
        sys.exit("\n!!! %s\nCheck and update your settings in: %s" % (err, scrapydweb_settings_py_path))

    if app.config.get('ENABLE_CACHE', True):
        caching_pid = init_caching(app.config, main_pid)
    else:
        caching_pid = None

    # https://stackoverflow.com/questions/34164464/flask-decorate-every-route-at-once
    @app.before_request
    def require_login():
        if app.config.get('ENABLE_AUTH', False):
            auth = request.authorization
            USERNAME = str(app.config.get('USERNAME', ''))  # May be 0 from config file
            PASSWORD = str(app.config.get('PASSWORD', ''))
            if not auth or not (auth.username == USERNAME and auth.password == PASSWORD):
                return authenticate()

    # Should be commented out for released version
    # https://stackoverflow.com/questions/34066804/disabling-caching-in-flask
    # @app.after_request
    # def add_header(r):
        # r.headers['Pragma'] = 'no-cache'
        # r.headers['Expires'] = '0'
        # r.headers['Cache-Control'] = 'public, max-age=0'
        # return r

    @app.context_processor
    def inject_variable():
        return dict(
            main_pid=main_pid,
            caching_pid=caching_pid,
            CHECK_LATEST_VERSION_FREQ=100,
            scrapydweb_settings_py_path=scrapydweb_settings_py_path,
        )

    printf("Visit ScrapydWeb at http://127.0.0.1:{port} or http://{bind}:{port}".format(
        bind='IP-OF-CURRENT-HOST', port=app.config['SCRAPYDWEB_PORT']))

    # site-packages/flask/app.py
    # def run(self, host=None, port=None, debug=None, load_dotenv=True, **options):
    # Threaded mode is enabled by default.
    app.run(host=app.config['SCRAPYDWEB_BIND'], port=app.config['SCRAPYDWEB_PORT'])  # , debug=True)


def load_custom_config(config):
    global scrapydweb_settings_py_path

    path = find_scrapydweb_settings_py(SCRAPYDWEB_SETTINGS_PY, os.getcwd())

    print('*' * 100)
    if path:
        scrapydweb_settings_py_path = path
        printf("Overriding custom settings from %s" % scrapydweb_settings_py_path, warn=True)
        config.from_pyfile(scrapydweb_settings_py_path)
    else:
        try:
            os.remove(LAST_CHECK_UPDATE)
        except:
            pass

        try:
            copyfile(os.path.join(CWD, 'default_settings.py'), scrapydweb_settings_py_path)
        except:
            sys.exit("Please copy the 'default_settings.py' file from above path to current working directory,\n"
                     "and rename it to '%s'.\n"
                     "Then add your SCRAPYD_SERVERS in the file and restart scrapydweb." % SCRAPYDWEB_SETTINGS_PY)
        else:
            sys.exit("The config file '%s' has been copied to current working directory.\n"
                     "Please add your SCRAPYD_SERVERS in the file and restart scrapydweb." % SCRAPYDWEB_SETTINGS_PY)


def parse_args(config):
    parser = argparse.ArgumentParser(description='ScrapydWeb -- %s' % __description__)

    SCRAPYDWEB_BIND = config.get('SCRAPYDWEB_BIND', '0.0.0.0')
    parser.add_argument(
        '-b', '--bind',
        default=SCRAPYDWEB_BIND,
        help=("current: %s, note that setting 0.0.0.0 or IP-OF-CURRENT-HOST makes ScrapydWeb server "
              "visible externally, otherwise, type '-b 127.0.0.1'") % SCRAPYDWEB_BIND
    )

    SCRAPYDWEB_PORT = config.get('SCRAPYDWEB_PORT', 5000)
    parser.add_argument(
        '-p', '--port',
        default=SCRAPYDWEB_PORT,
        help="current: %s, the port which ScrapydWeb would run on" % SCRAPYDWEB_PORT
    )

    SCRAPYD_SERVERS = config.get('SCRAPYD_SERVERS', []) or ['127.0.0.1:6800']
    parser.add_argument(
        '-ss', '--scrapyd_server',
        action='append',
        help=("current: %s, type '-ss 127.0.0.1 -ss username:password@192.168.123.123:6801#group' "
              "to set up more than one Scrapyd server to control. ") % SCRAPYD_SERVERS
    )

    ENABLE_AUTH = config.get('ENABLE_AUTH', False)
    parser.add_argument(
        '-da', '--disable_auth',
        action='store_true',
        help="current: ENABLE_AUTH = %s, append '--disable_auth' to disable basic auth for web UI" % ENABLE_AUTH
    )

    ENABLE_CACHE = config.get('ENABLE_CACHE', True)
    parser.add_argument(
        '-dc', '--disable_cache',
        action='store_true',
        help=("current: ENABLE_CACHE = %s, append '--disable_cache' to disable caching HTML for Log and Stats page "
              "in the background periodically") % ENABLE_CACHE
    )

    DELETE_CACHE = config.get('DELETE_CACHE', False)
    parser.add_argument(
        '-del', '--delete_cache',
        action='store_true',
        help=("current: DELETE_CACHE = %s, append '--delete_cache' to delete cached HTML files "
              "of Log and Stats page at startup" % DELETE_CACHE)
    )

    ENABLE_EMAIL = config.get('ENABLE_EMAIL', False)
    parser.add_argument(
        '-de', '--disable_email',
        action='store_true',
        help="current: ENABLE_EMAIL = %s, append '--disable_email' to disable email notice" % ENABLE_EMAIL
    )

    DEBUG = config.get('DEBUG', False)
    parser.add_argument(
        '-d', '--debug',
        action='store_true',
        help=("current: DEBUG = %s, append '--debug' to enable debug mode "
              "and the debugger would be available in the browser") % DEBUG
    )

    VERBOSE = config.get('VERBOSE', False)
    parser.add_argument(
        '-v', '--verbose',
        action='store_true',
        help=("current: VERBOSE = %s, append '--verbose' to set logging level to DEBUG "
              "for getting more information about how ScrapydWeb works") % VERBOSE
    )

    return parser.parse_args()


def update_app_config(config, args):
    printf("Reading settings from command line: %s" % args)

    config.update(dict(
        SCRAPYDWEB_BIND=args.bind,
        SCRAPYDWEB_PORT=args.port,
    ))

    # scrapyd_server would be None if -ss not passed in
    SCRAPYD_SERVERS = args.scrapyd_server or config.get('SCRAPYD_SERVERS', []) or ['127.0.0.1:6800']
    servers = []
    for idx, server in enumerate(SCRAPYD_SERVERS):
        if isinstance(server, tuple):
            assert len(server) == 5, ("Scrapyd server should be a tuple with 5 elements, "
                                      "current value: %s" % str(server))
            usr, psw, ip, port, group = server
        else:
            usr, psw, ip, port, group = pattern_scrapyd_server.search(server.strip()).groups()
        ip = ip.strip() if ip and ip.strip() else '127.0.0.1'
        port = port.strip() if port and port.strip() else '6800'
        group = group.strip() if group and group.strip() else ''
        auth = (usr, psw) if usr and psw else None
        servers.append((group, ip, port, auth))

    def key(arg):
        group, ip, port, auth = arg
        parts = ip.split('.')
        parts = [('0' * (3 - len(part)) + part) for part in parts]
        return [group, '.'.join(parts), int(port)]

    servers = sorted(set(servers), key=key)
    check_scrapyd_connectivity(servers)

    config['SCRAPYD_SERVERS'] = ['%s:%s' % (ip, port) for group, ip, port, auth in servers]
    config['SCRAPYD_SERVERS_GROUPS'] = [group for group, ip, port, auth in servers]
    config['SCRAPYD_SERVERS_AUTHS'] = [auth for group, ip, port, auth in servers]

    # action='store_true': default False
    if args.disable_auth:
        config['ENABLE_AUTH'] = False
    if args.disable_cache:
        config['ENABLE_CACHE'] = False
    if args.delete_cache:
        config['DELETE_CACHE'] = True
    if args.disable_email:
        config['ENABLE_EMAIL'] = False
    if args.debug:
        config['DEBUG'] = True
    if args.verbose:
        config['VERBOSE'] = True


def check_scrapyd_connectivity(servers):
    printf("Checking connectivity of SCRAPYD_SERVERS")

    def check_connectivity(server):
        (group, ip, port, auth) = server
        try:
            r = requests.get('http://%s:%s' % (ip, port), auth=auth, timeout=3)
            assert r.status_code == 200
        except:
            return False
        else:
            return True

    # with ThreadPool(min(len(servers), 10)) as pool:  # Works in python 3.3 and up
        # results = pool.map(check_connectivity, servers)
    pool = ThreadPool(min(len(servers), 10))
    results = pool.map(check_connectivity, servers)
    pool.close()
    pool.join()

    print("Index {group:<20} {server:<21} Connectivity Auth".format(
          group='Group', server='Scrapyd IP:Port'))
    print('#' * 100)
    for idx, ((group, ip, port, auth), result) in enumerate(zip(servers, results), 1):
        print("{idx:_<5} {group:_<20} {server:_<22} {result:_<11} {auth}".format(
              idx=idx, group=group or 'None', server='%s:%s' % (ip, port), auth=auth, result=str(result)))
    print('#' * 100)

    if not any(results):
        sys.exit("\n!!! None of your SCRAPYD_SERVERS could be connected.\n"
                 "Check and update the SCRAPYD_SERVERS item in: %s" % scrapydweb_settings_py_path)


if __name__ == '__main__':
    main()



# coding: utf8
import re

from flask import render_template

from ..myview import MyView
from ..vars import pattern_directory, keys_directory


class ItemsView(MyView):
    methods = ['GET']

    def __init__(self):
        super(self.__class__, self).__init__()

        self.project = self.view_args['project']
        self.spider = self.view_args['spider']

        self.url = 'http://{}/items/{}{}'.format(self.SCRAPYD_SERVER,
                                                 '%s/' % self.project if self.project else '',
                                                 '%s/' % self.spider if self.spider else '')
        self.template = 'scrapydweb/items.html'

        self.text = ''

    def dispatch_request(self, **kwargs):
        status_code, self.text = self.make_request(self.url, api=False, auth=self.AUTH)
        if status_code != 200 or not re.search(r'Directory listing for /items/', self.text):
            kwargs = dict(
                node=self.node,
                url=self.url,
                status_code=status_code,
                text=self.text,
                message='Check out https://scrapyd.readthedocs.io/en/latest/config.html#items-dir for help'
            )
            return render_template(self.template_fail, **kwargs)

        return self.generate_response()

    def generate_response(self):
        rows = [dict(zip(keys_directory, row)) for row in pattern_directory.findall(self.text)]

        for row in rows:
            # <a href="demo/">demo/</a>
            # <a href="test/">test/</a>
            # <a href="2018-10-09_225255.jl">2018-10-09_225255.jl</a>
            if self.project and self.spider:
                m = re.search(r'>(.*?)<', row['filename'])
                filename = m.group(1)
                row['filename'] = u'<a class="link" target="_blank" href="{}">{}</a>'.format(
                                   self.url + filename, filename)
            else:
                row['filename'] = re.sub(r'href=', 'class="link" href=', row['filename'])

        kwargs = dict(
            node=self.node,
            project=self.project,
            spider=self.spider,
            url=self.url,
            rows=rows
        )

        return render_template(self.template, **kwargs)



# coding: utf8
import sys
import time
import re
import json
import logging

import requests
from requests.adapters import HTTPAdapter
from requests.auth import _basic_auth_str
from flask import request, url_for, g
from flask import current_app as app
from flask.views import View

from .__version__ import __version__, __url__
from .vars import DEMO_PROJECTS_PATH, ALLOWED_SCRAPYD_LOG_EXTENSIONS, EMAIL_TRIGGER_KEYS, DEFAULT_LATEST_VERSION
from .utils.utils import json_dumps


session = requests.Session()
session.mount('http://', HTTPAdapter(pool_connections=1000, pool_maxsize=1000))
session.mount('https://', HTTPAdapter(pool_connections=1000, pool_maxsize=1000))


class MyView(View):
    methods = ['GET', 'POST']

    def __init__(self, *args, **kwargs):
        self.logger = logging.getLogger(self.__class__.__name__)

        # System
        self.DEBUG = app.config.get('DEBUG', False)
        self.VERBOSE = app.config.get('VERBOSE', False)

        _level = logging.DEBUG if self.VERBOSE else logging.WARNING
        self.logger.setLevel(_level)
        logging.getLogger("requests").setLevel(_level)
        logging.getLogger("urllib3").setLevel(_level)

        if request.form:
            self.logger.debug(self.json_dumps(request.form))

        # ScrapydWeb
        self.SCRAPYDWEB_BIND = app.config.get('SCRAPYDWEB_BIND', '0.0.0.0')
        self.SCRAPYDWEB_PORT = app.config.get('SCRAPYDWEB_PORT', 5000)

        self.ENABLE_AUTH = app.config.get('ENABLE_AUTH', False)
        self.USERNAME = app.config.get('USERNAME', '')
        self.PASSWORD = app.config.get('PASSWORD', '')

        # Scrapy
        self.SCRAPY_PROJECTS_DIR = app.config.get('SCRAPY_PROJECTS_DIR', '') or DEMO_PROJECTS_PATH

        # Scrapyd
        self.SCRAPYD_SERVERS = app.config.get('SCRAPYD_SERVERS', []) or ['127.0.0.1:6800']
        self.SCRAPYD_SERVERS_GROUPS = app.config.get('SCRAPYD_SERVERS_GROUPS', []) or ['']
        self.SCRAPYD_SERVERS_AUTHS = app.config.get('SCRAPYD_SERVERS_AUTHS', []) or [None]

        self.SCRAPYD_LOGS_DIR = app.config.get('SCRAPYD_LOGS_DIR', '')
        self.SCRAPYD_LOG_EXTENSIONS = (app.config.get('SCRAPYD_LOG_EXTENSIONS', [])
                                       or ALLOWED_SCRAPYD_LOG_EXTENSIONS)

        # Page Display
        self.SHOW_SCRAPYD_ITEMS = app.config.get('SHOW_SCRAPYD_ITEMS', True)
        self.SHOW_DASHBOARD_JOB_COLUMN = app.config.get('SHOW_DASHBOARD_JOB_COLUMN', False)
        self.DASHBOARD_RELOAD_INTERVAL = app.config.get('DASHBOARD_RELOAD_INTERVAL', 300)
        self.DAEMONSTATUS_REFRESH_INTERVAL = app.config.get('DAEMONSTATUS_REFRESH_INTERVAL', 10)
        self.LAST_LOG_ALERT_SECONDS = app.config.get('LAST_LOG_ALERT_SECONDS', 60)  # Not in default_settings.py

        # HTML Caching
        self.ENABLE_CACHE = app.config.get('ENABLE_CACHE', True)
        self.CACHE_ROUND_INTERVAL = app.config.get('CACHE_ROUND_INTERVAL', 300)
        self.CACHE_REQUEST_INTERVAL = app.config.get('CACHE_REQUEST_INTERVAL', 10)
        self.DELETE_CACHE = app.config.get('DELETE_CACHE', False)

        # Email Notice
        self.ENABLE_EMAIL = app.config.get('ENABLE_EMAIL', False)
        self.SMTP_SERVER = app.config.get('SMTP_SERVER', '')
        self.SMTP_PORT = app.config.get('SMTP_PORT', 0)
        self.SMTP_OVER_SSL = app.config.get('SMTP_OVER_SSL', False)
        self.SMTP_CONNECTION_TIMEOUT = app.config.get('SMTP_CONNECTION_TIMEOUT', 10)
        self.FROM_ADDR = app.config.get('FROM_ADDR', '')
        self.EMAIL_PASSWORD = app.config.get('EMAIL_PASSWORD', '')
        self.TO_ADDRS = app.config.get('TO_ADDRS', [])

        self.EMAIL_KWARGS = dict(
            smtp_server=self.SMTP_SERVER,
            smtp_port=self.SMTP_PORT,
            smtp_over_ssl=self.SMTP_OVER_SSL,
            smtp_connection_timeout=self.SMTP_CONNECTION_TIMEOUT,
            from_addr=self.FROM_ADDR,
            email_password=self.EMAIL_PASSWORD,
            to_addrs=self.TO_ADDRS,
            subject='subject',
            content='content'
        )

        self.EMAIL_WORKING_DAYS = app.config.get('EMAIL_WORKING_DAYS', [])
        self.EMAIL_WORKING_HOURS = app.config.get('EMAIL_WORKING_HOURS', [])
        self.ON_JOB_RUNNING_INTERVAL = app.config.get('ON_JOB_RUNNING_INTERVAL', 0)
        self.ON_JOB_FINISHED = app.config.get('ON_JOB_FINISHED', False)
        # ['CRITICAL', 'ERROR', 'WARNING', 'REDIRECT', 'RETRY', 'IGNORE']
        for key in EMAIL_TRIGGER_KEYS:
            setattr(self, 'LOG_%s_THRESHOLD' % key, app.config.get('LOG_%s_THRESHOLD' % key, 0))
            setattr(self, 'LOG_%s_TRIGGER_STOP' % key, app.config.get('LOG_%s_TRIGGER_STOP' % key, False))
            setattr(self, 'LOG_%s_TRIGGER_FORCESTOP' % key, app.config.get('LOG_%s_TRIGGER_FORCESTOP' % key, False))

        # Other attributes NOT from config
        self.view_args = request.view_args
        self.node = self.view_args['node']
        self.SCRAPYD_SERVER = self.SCRAPYD_SERVERS[self.node - 1]
        self.GROUP = self.SCRAPYD_SERVERS_GROUPS[self.node - 1]
        self.AUTH = self.SCRAPYD_SERVERS_AUTHS[self.node - 1]

        ua = request.headers.get('User-Agent', '')
        m_mobile = re.search(r'Android|webOS|iPad|iPhone|iPod|BlackBerry|IEMobile|Opera Mini', ua, re.I)
        self.IS_MOBILE = True if m_mobile else False

        m_ipad = re.search(r'iPad', ua, re.I)
        self.IS_IPAD = True if m_ipad else False

        # http://werkzeug.pocoo.org/docs/0.14/utils/#module-werkzeug.useragents
        # /site-packages/werkzeug/useragents.py
        browser = request.user_agent.browser or ''  # lib requests GET: None
        m_edge = re.search(r'Edge', ua, re.I)
        self.IS_IE_EDGE = True if (browser == 'msie' or m_edge) else False

        self.IS_MOBILEUI = True if request.args.get('ui', '') == 'mobile' else False
        self.UI = 'mobile' if self.IS_MOBILEUI else None
        self.GET = True if request.method == 'GET' else False
        self.POST = True if request.method == 'POST' else False

        # In log.py: pass self.IS_MOBILE to render_template() to override g.IS_MOBILE
        g.IS_MOBILE = self.IS_MOBILE  # lifetime: every single request

        self.template_fail = 'scrapydweb/fail_mobileui.html' if self.IS_MOBILEUI else 'scrapydweb/fail.html'

        self.update_g()
        self.inject_variable(version='v100rc2')

    def update_g(self):
        # For base.html and base_mobileui.html
        g.url_menu_overview = url_for('overview', node=self.node)
        g.url_menu_dashboard = url_for('dashboard', node=self.node)
        g.url_menu_deploy = url_for('deploy.deploy', node=self.node)
        g.url_menu_schedule = url_for('schedule.schedule', node=self.node)
        g.url_menu_manage = url_for('manage', node=self.node)
        g.url_menu_items = url_for('items', node=self.node)
        g.url_menu_logs = url_for('logs', node=self.node)
        g.url_menu_parse = url_for('parse.upload', node=self.node)
        g.url_menu_settings = url_for('settings', node=self.node)
        g.url_menu_mobileui = url_for('index', node=self.node, ui='mobile')

        g.url_daemonstatus = url_for('api', node=self.node, opt='daemonstatus')

    def inject_variable(self, version):
        @app.context_processor
        def inject_variable():
            return dict(
                SCRAPYD_SERVERS=self.SCRAPYD_SERVERS,
                SCRAPYD_SERVERS_AMOUNT=len(self.SCRAPYD_SERVERS),
                SCRAPYD_SERVERS_GROUPS=self.SCRAPYD_SERVERS_GROUPS,
                SCRAPYD_SERVERS_AUTHS=self.SCRAPYD_SERVERS_AUTHS,
                PYTHON_VERSION='.'.join([str(n) for n in sys.version_info[:3]]),
                SCRAPYDWEB_VERSION=__version__,
                # CHECK_LATEST_VERSION_FREQ=100,  #Would override inject_variable() in test_page.py
                GITHUB_URL=__url__,
                DEFAULT_LATEST_VERSION=DEFAULT_LATEST_VERSION,
                SHOW_SCRAPYD_ITEMS=self.SHOW_SCRAPYD_ITEMS,
                DAEMONSTATUS_REFRESH_INTERVAL=self.DAEMONSTATUS_REFRESH_INTERVAL,
                REQUIRE_LOGIN=True if app.config.get('ENABLE_AUTH', False) else False,

                static_css_common=url_for('static', filename='%s/css/common.css' % version),
                static_css_dropdown=url_for('static', filename='%s/css/dropdown.css' % version),
                static_css_dropdown_mobileui=url_for('static', filename='%s/css/dropdown_mobileui.css' % version),
                static_css_icon_upload_icon_right=url_for('static',
                                                          filename='%s/css/icon_upload_icon_right.css' % version),
                static_css_multinode=url_for('static', filename='%s/css/multinode.css' % version),
                static_css_stacktable=url_for('static', filename='%s/css/stacktable.css' % version),
                static_css_stats=url_for('static', filename='%s/css/stats.css' % version),
                static_css_style=url_for('static', filename='%s/css/style.css' % version),
                static_css_style_mobileui=url_for('static', filename='%s/css/style_mobileui.css' % version),
                static_css_utf8=url_for('static', filename='%s/css/utf8.css' % version),
                static_css_utf8_mobileui=url_for('static', filename='%s/css/utf8_mobileui.css' % version),

                static_css_element_ui_index=url_for('static',
                                                    filename='%s/element-ui@2.4.6/lib/theme-chalk/index.css' % version),
                static_js_element_ui_index=url_for('static', filename='%s/element-ui@2.4.6/lib/index.js' % version),

                static_js_common=url_for('static', filename='%s/js/common.js' % version),
                static_js_echarts_min=url_for('static', filename='%s/js/echarts.min.js' % version),
                static_js_icons_menu=url_for('static', filename='%s/js/icons_menu.js' % version),
                # static_js_github_buttons_html=url_for('static', filename='%s/js/github_buttons.html' % version),
                static_js_github_buttons=url_for('static', filename='%s/js/github_buttons.js' % version),
                static_js_jquery_min=url_for('static', filename='%s/js/jquery.min.js' % version),
                static_js_multinode=url_for('static', filename='%s/js/multinode.js' % version),
                static_js_stacktable=url_for('static', filename='%s/js/stacktable.js' % version),
                static_js_stats=url_for('static', filename='%s/js/stats.js' % version),
                static_js_vue_min=url_for('static', filename='%s/js/vue.min.js' % version),

                static_icon=url_for('static', filename='%s/icon/fav.ico' % version),
                static_icon_shortcut=url_for('static', filename='%s/icon/fav.ico' % version),
                static_icon_apple_touch=url_for('static', filename='%s/icon/spiderman.png' % version),

                # For base.html and base_mobileui.html
                # url_menu_XXX might be update by caching subprocess, use update_g() instead
                # url_menu_overview=url_for('overview', node=self.node),
                url_dashboard_list=[url_for('dashboard', node=n, ui=self.UI)
                                    for n in range(1, len(self.SCRAPYD_SERVERS)+1)],
            )

    def get_selected_nodes(self):
        selected_nodes = []
        for i in range(1, len(self.SCRAPYD_SERVERS) + 1):
            if request.form.get(str(i)) == 'on':
                selected_nodes.append(i)
        return selected_nodes

    def get_response_from_view(self, url):
        # https://stackoverflow.com/a/21342070/10517783  How do I call one Flask view from another one?
        # https://stackoverflow.com/a/30250045/10517783
        # python - Flask test_client() doesn't have request.authorization with pytest
        client = app.test_client()
        if self.ENABLE_AUTH:
            headers = {'Authorization': _basic_auth_str(self.USERNAME, self.PASSWORD)}
        else:
            headers = {}
        response = client.get(url, headers=headers)
        return response.get_data(as_text=True)

    @staticmethod
    def get_now_string():
        return time.strftime('%Y-%m-%dT%H_%M_%S')

    @staticmethod
    def json_dumps(obj, sort_keys=True):
        return json_dumps(obj, sort_keys=sort_keys)

    def make_request(self, url, data=None, timeout=60, api=True, auth=None):
        """
        :param url: url to make request
        :param data: None or a dict object to post
        :param timeout: timeout when making request, in seconds
        :param api: return a dict object if set True, else text
        :param auth: None or (username, password) for basic auth
        """
        try:
            if 'addversion.json' in url and data:
                self.logger.debug('>>>>> POST %s' % url)
                self.logger.debug(json_dumps(dict(project=data['project'], version=data['version'],
                                                  egg="%s bytes binary egg file" % len(data['egg']))))
            else:
                self.logger.debug('>>>>> %s %s' % ('POST' if data else 'GET', url))
                if data:
                    self.logger.debug(json_dumps(data))

            if data:
                r = session.post(url, data=data, timeout=timeout, auth=auth)
            else:
                r = session.get(url, timeout=timeout, auth=auth)
            r.encoding = 'utf8'
        except Exception as err:
            self.logger.error('!!!!! %s %s' % (err.__class__.__name__, err))
            if api:
                r_json = dict(url=url, auth=auth, status_code=-1,
                              status='error', message=str(err), when=time.ctime())
                return -1, r_json
            else:
                return -1, str(err)
        else:
            if api:
                r_json = {}
                try:
                    r_json = r.json()
                except json.JSONDecodeError:  # listprojects would get 502 html when Scrapyd server reboots
                    r_json = {'status': 'error', 'message': r.text}
                finally:
                    r_json.update(dict(url=url, auth=auth, status_code=r.status_code, when=time.ctime()))
                    if r.status_code != 200 or r_json.get('status') != 'ok':
                        self.logger.error('!!!!! %s %s' % (r.status_code, url))
                        self.logger.error(json_dumps(r_json))
                    elif not url.endswith('daemonstatus.json'):
                        self.logger.debug('<<<<< %s %s' % (r.status_code, url))
                        self.logger.debug(json_dumps(r_json))

                    return r.status_code, r_json
            else:
                if r.status_code == 200:
                    front = r.text[:min(100, len(r.text))].replace('\n', '')
                    back = r.text[-min(100, len(r.text)):].replace('\n', '')
                    self.logger.debug('<<<<< %s %s\n...%s' % (r.status_code, front, back))
                else:
                    self.logger.error('!!!!! %s %s' % (r.status_code, r.text))

                return r.status_code, r.text



# coding: utf8
import os
import sys
import platform
import time
import re
import json

import requests
from requests.adapters import HTTPAdapter
try:
    from psutil import pid_exists
except ImportError:
    PSUTIL_AVAILABLE = False
else:
    PSUTIL_AVAILABLE = True


verbose = True
IN_WINDOWS = True if platform.system() == 'Windows' else False
pattern_jobs = re.compile(r"""<tr>
                        <td>(?P<Project>.*?)</td>
                        <td>(?P<Spider>.*?)</td>
                        <td>(?P<Job>.*?)</td>
                        (?:<td>(?P<PID>.*?)</td>)?
                        (?:<td>(?P<Start>.*?)</td>)?
                        (?:<td>(?P<Runtime>.*?)</td>)?
                        (?:<td>(?P<Finish>.*?)</td>)?
                        (?:<td>(?P<Log>.*?)</td>)?
                        (?:<td>(?P<Items>.*?)</td>)?
                        </tr>
                    """, re.X)
keys_jobs = ['project', 'spider', 'job', 'pid', 'start', 'runtime', 'finish', 'log', 'items']


def printf(value, warn=False):
    if not verbose and not warn:
        return
    prefix = "!!! " if warn else ""
    print("%sHTML caching %s" % (prefix, value))


# https://stackoverflow.com/questions/568271/how-to-check-if-there-exists-a-process-with-a-given-pid-in-python
def check_pid(pid):
    """ Check For the existence of a unix pid. """
    try:
        os.kill(pid, 0)
    except OSError:
        return False
    else:
        return True


class RefreshCache(object):

    def __init__(self, scrapydweb_bind, scrapydweb_port,
                 node, SCRAPYD_SERVER, auth,
                 session, timeout, cache_request_interval,
                 ignore_finished_bool_list, finished_jobs_dict):
        self.scrapydweb_bind = scrapydweb_bind
        self.scrapydweb_port = scrapydweb_port
        self.node = node
        self.SCRAPYD_SERVER = SCRAPYD_SERVER
        self.auth = tuple(auth) if auth else auth  # TypeError: 'list' object is not callable
        self.session = session
        self.timeout = timeout
        self.cache_request_interval = cache_request_interval
        self.ignore_finished_bool_list = ignore_finished_bool_list
        self.finished_jobs_dict = finished_jobs_dict

        self.url_jobs = 'http://%s/jobs' % self.SCRAPYD_SERVER
        self.running_jobs = []
        self.finished_jobs_set = set()
        self.finished_jobs = []

        self.url_stats = ('http://{scrapydweb_bind}:{scrapydweb_port}/'
                          '{node}/log/{opt}/{project}/{spider}/{job}/?job_finished={job_finished}')

    def main(self):
        self.fetch_jobs()

        for job_tuple in self.running_jobs + self.finished_jobs:
            self.fetch_stats(job_tuple)

    def fetch_jobs(self):
        try:
            r = self.session.get(self.url_jobs, auth=self.auth, timeout=self.timeout)
            assert r.status_code == 200, "got status_code: %s" % r.status_code
        except Exception as err:
            # import traceback
            # print(traceback.format_exc())
            printf("request jobs failed: %s" % err, warn=True)
        else:
            printf("request jobs got (%s) %s bytes: %s" % (r.status_code, len(r.content), self.url_jobs))
            rows = [dict(zip(keys_jobs, row)) for row in pattern_jobs.findall(r.text)]
            for row in rows:
                job_tuple = (row['project'], row['spider'], row['job'])
                if row['pid']:
                    self.running_jobs.append(job_tuple)
                elif row['finish']:
                    self.finished_jobs_set.add(job_tuple)

            self.update_finished_jobs()
        finally:
            time.sleep(self.cache_request_interval)

    def update_finished_jobs(self):
        finished_jobs_set_previous = self.finished_jobs_dict.setdefault(self.node, set())
        # set([2,3]).difference(set([1,2])) => {3}
        finished_jobs_set_new_added = self.finished_jobs_set.difference(finished_jobs_set_previous)
        self.finished_jobs_dict[self.node] = self.finished_jobs_set
        # print(len(self.finished_jobs_dict[self.node]))

        ignore = self.ignore_finished_bool_list[self.node-1]
        for job_tuple in finished_jobs_set_new_added:
            if ignore:
                printf('ignore finished job from node %s: %s' % (self.node, job_tuple))
            else:
                self.finished_jobs.append(job_tuple)
        if ignore:
            self.ignore_finished_bool_list[self.node-1] = False

    def fetch_stats(self, job_tuple):
        (project, spider, job) = job_tuple
        job_finished = True if job_tuple in self.finished_jobs else ''
        kwargs = dict(
            scrapydweb_bind=self.scrapydweb_bind,
            scrapydweb_port=self.scrapydweb_port,
            node=self.node,
            opt='stats',
            project=project,
            spider=spider,
            job=job,
            job_finished=job_finished
        )
        # http://127.0.0.1:5000/log/utf8/proxy/test/55f1f388a7ae11e8b9b114dda9e91c2f/
        url = self.url_stats.format(**kwargs)
        try:
            # Make POST request to avoid using cached HTML, see log.py
            r = session.post(url, timeout=self.timeout)
            assert r.status_code == 200, "got status_code %s" % r.status_code
        except Exception as err:
            printf("request stats page failed: %s" % err, warn=True)
            if job_finished:
                self.finished_jobs_dict[self.node].remove(job_tuple)
        else:
            printf("request stats page got (%s) %s bytes: %s" % (r.status_code, len(r.content), url))
        finally:
            time.sleep(self.cache_request_interval)


def main():
    while True:
        if (PSUTIL_AVAILABLE and not pid_exists(main_pid)) or (not IN_WINDOWS and not check_pid(main_pid)):
            sys.exit("!!! Caching subprocess %s exit: main_pid %s NOT exists" % (caching_pid, main_pid))
        start_time = time.time()
        try:
            for node, SCRAPYD_SERVER in enumerate(SCRAPYD_SERVERS, 1):
                updater = RefreshCache(scrapydweb_bind, scrapydweb_port,
                                       node, SCRAPYD_SERVER, SCRAPYD_SERVERS_AUTHS[node-1],
                                       session, timeout, cache_request_interval,
                                       ignore_finished_bool_list, finished_jobs_dict)
                updater.main()

            printf("done at %s" % time.ctime())
            printf("cost %s seconds" % int(time.time() - start_time))
            printf("sleep %s seconds" % cache_round_interval)
            time.sleep(cache_round_interval)
        except KeyboardInterrupt:
            sys.exit("!!! Caching subprocess %s cancelled by KeyboardInterrupt" % caching_pid)
        except Exception as err:
            printf("error: %s" % err, warn=True)


if __name__ == '__main__':
    (main_pid,
     scrapydweb_bind, scrapydweb_port, username, password,
     SCRAPYD_SERVERS, SCRAPYD_SERVERS_AUTHS,
     cache_round_interval, cache_request_interval,
     verbose) = sys.argv[1:]

    main_pid = int(main_pid)
    SCRAPYD_SERVERS = json.loads(SCRAPYD_SERVERS)
    SCRAPYD_SERVERS_AUTHS = json.loads(SCRAPYD_SERVERS_AUTHS)
    cache_round_interval = int(cache_round_interval)
    cache_request_interval = int(cache_request_interval)
    verbose = True if verbose == 'True' else False

    session = requests.Session()
    session.mount('http://', HTTPAdapter(pool_connections=1000, pool_maxsize=1000))
    session.mount('https://', HTTPAdapter(pool_connections=1000, pool_maxsize=1000))
    if username and password:
        session.auth = (username, password)
    timeout = 60

    caching_pid = os.getpid()
    ignore_finished_bool_list = [True] * len(SCRAPYD_SERVERS)
    finished_jobs_dict = {}

    main()



# coding: utf8
import os
import io

from setuptools import find_packages, setup


CWD = os.path.dirname(os.path.abspath(__file__))

about = {}
with open(os.path.join(CWD, 'scrapydweb', '__version__.py')) as f:
    exec(f.read(), about)

with io.open("README.md", 'r', encoding='utf8') as f:
    long_description = f.read()


setup(
    name=about['__title__'],
    version=about['__version__'],
    author=about['__author__'],
    author_email=about['__author_email__'],
    url=about['__url__'],
    license=about['__license__'],
    description=about['__description__'],

    long_description=long_description,
    long_description_content_type="text/markdown",

    packages=find_packages(exclude=("tests", )),
    include_package_data=True,
    zip_safe=False,
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*",
    install_requires=[
        "flask >= 1.0.2",
        "flask-compress >= 1.4.0",
        "requests",
        "setuptools",
    ],

    entry_points={
        "console_scripts": {
            "scrapydweb = scrapydweb.run:main"
        },
    },

    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7"
    ],
)
